import os
import shutil
import tempfile
from typing import (
    Optional,
    TYPE_CHECKING,
)

import tool_shed.repository_types.util as rt_util
from galaxy.tool_shed.util.hg_util import clone_repository
from galaxy.util.compression_utils import CompressedFile
from tool_shed.dependencies.attribute_handlers import (
    RepositoryDependencyAttributeHandler,
    ToolDependencyAttributeHandler,
)
from tool_shed.util import xml_util
from tool_shed.util.commit_util import (
    ChangeResponseT,
    check_archive,
    handle_directory_changes,
)

if TYPE_CHECKING:
    from tool_shed.context import ProvidesRepositoriesContext
    from tool_shed.webapp.model import Repository


def upload_tar(
    trans: "ProvidesRepositoriesContext",
    username: str,
    repository: "Repository",
    uploaded_file,
    commit_message: str,
    dry_run: bool = False,
    remove_repo_files_not_in_tar: bool = True,
    new_repo_alert: bool = False,
    rdah: Optional[RepositoryDependencyAttributeHandler] = None,
    tdah: Optional[ToolDependencyAttributeHandler] = None,
) -> ChangeResponseT:
    host = trans.repositories_hostname
    app = trans.app
    tar = CompressedFile.open_tar(uploaded_file)
    rdah = rdah or RepositoryDependencyAttributeHandler(trans, unpopulate=False)
    tdah = tdah or ToolDependencyAttributeHandler(trans, unpopulate=False)
    # Upload a tar archive of files.
    undesirable_dirs_removed = 0
    undesirable_files_removed = 0
    check_results = check_archive(repository, tar)
    if check_results.invalid:
        tar.close()
        try:
            uploaded_file.close()
        except AttributeError:
            pass
        message = "{} Invalid paths were: {}".format(
            " ".join(check_results.errors), ", ".join([i.name for i in check_results.invalid])
        )
        return False, message, [], "", undesirable_dirs_removed, undesirable_files_removed
    else:
        repo_dir = repository.repo_path(app)
        if dry_run:
            full_path = tempfile.mkdtemp()
            clone_repository(repo_dir, full_path)
        else:
            full_path = os.path.abspath(repo_dir)
        undesirable_files_removed = len(check_results.undesirable_files)
        undesirable_dirs_removed = len(check_results.undesirable_dirs)
        filenames_in_archive = [ti.name for ti in check_results.valid]
        # Extract the uploaded tar to the load_point within the repository hierarchy.
        tar.extractall(path=full_path, members=check_results.valid)
        tar.close()
        try:
            uploaded_file.close()
        except AttributeError:
            pass
        for filename in filenames_in_archive:
            uploaded_file_name = os.path.join(full_path, filename)
            if os.path.split(uploaded_file_name)[-1] == rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME:
                # Inspect the contents of the file to see if toolshed or changeset_revision attributes
                # are missing and if so, set them appropriately.
                altered, root_elem, error_message = rdah.handle_tag_attributes(uploaded_file_name)
                if error_message:
                    return False, error_message, [], "", 0, 0
                elif altered:
                    tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                    shutil.move(tmp_filename, uploaded_file_name)
            elif os.path.split(uploaded_file_name)[-1] == rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME:
                # Inspect the contents of the file to see if toolshed or changeset_revision
                # attributes are missing and if so, set them appropriately.
                altered, root_elem, error_message = tdah.handle_tag_attributes(uploaded_file_name)
                if error_message:
                    return False, error_message, [], "", 0, 0
                if altered:
                    tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                    shutil.move(tmp_filename, uploaded_file_name)
        return handle_directory_changes(
            app,
            host,
            username,
            repository,
            full_path,
            filenames_in_archive,
            remove_repo_files_not_in_tar,
            new_repo_alert,
            commit_message,
            undesirable_dirs_removed,
            undesirable_files_removed,
            repo_path=full_path,
        )



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

column_repository_name = "column_maker_0150"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0150"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

category_name = "Test 0150 Simple Prior Installation"
category_description = "Test 0150 Simple Prior Installation"

"""
Create column_maker and convert_chars.

Column maker repository dependency:
<repository toolshed="self.url" name="convert_chars" owner="test" changeset_revision="c3041382815c" prior_installation_required="True" />

Verify display.

Galaxy side:

Install column_maker.
Verify that convert_chars was installed first, contrary to the ordering that would be present without prior_installation_required.
"""

running_standalone = False


class TestSimplePriorInstallation(ShedTwillTestCase):
    """Test features related to datatype converters."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_convert_repository(self):
        """Create and populate convert_chars_0150."""
        global running_standalone
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            running_standalone = True
            self.commit_tar_to_repository(
                repository,
                "convert_chars/convert_chars.tar",
                commit_message="Uploaded convert_chars tarball.",
            )

    def test_0010_create_column_repository(self):
        """Create and populate convert_chars_0150."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if running_standalone:
            self.commit_tar_to_repository(
                repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0015_create_repository_dependency(self):
        """Create a repository dependency specifying convert_chars.

        Column maker repository dependency:
            <repository toolshed="self.url" name="convert_chars" owner="test" changeset_revision="<tip>" prior_installation_required="True" />
        """
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        if running_standalone:
            dependency_xml_path = self.generate_temp_path("test_1150", additional_paths=["column"])
            convert_tuple = (
                self.url,
                convert_repository.name,
                convert_repository.owner,
                self.get_repository_tip(convert_repository),
            )
            self.create_repository_dependency(
                repository=column_repository,
                repository_tuples=[convert_tuple],
                filepath=dependency_xml_path,
                prior_installation_required=True,
            )

    def test_0020_verify_repository_dependency(self):
        """Verify that the previously generated repositiory dependency displays correctly."""
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        self.check_repository_dependency(
            repository=column_repository,
            depends_on_repository=convert_repository,
            depends_on_changeset_revision=None,
            changeset_revision=None,
        )

    def test_0025_install_column_repository(self):
        """Install column_maker_0150."""
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        preview_strings_displayed = ["column_maker_0150", self.get_repository_tip(column_repository)]
        self._install_repository(
            column_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            preview_strings_displayed=preview_strings_displayed,
        )

    def test_0030_verify_installation_order(self):
        """Verify that convert_chars_0150 was installed before column_maker_0150."""
        column_repository = self._get_installed_repository_by_name_owner(
            column_repository_name, common.test_user_1_name
        )
        convert_repository = self._get_installed_repository_by_name_owner(
            convert_repository_name, common.test_user_1_name
        )
        # Column maker was selected for installation, so convert chars should have been installed first, as reflected by the update_time field.
        assert (
            column_repository.update_time > convert_repository.update_time
        ), "Error: column_maker_0150 shows an earlier update time than convert_chars_0150"



import logging
from typing import (
    Any,
    Dict,
    List,
)

import tool_shed.util.shed_util_common as suc
import tool_shed_client.schema
from galaxy import (
    util,
    web,
)
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
    require_admin,
)
from galaxy.webapps.galaxy.api import depends
from tool_shed.managers.categories import CategoryManager
from tool_shed.managers.repositories import repositories_by_category
from tool_shed.webapp.model import Category
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class CategoriesController(BaseShedAPIController):
    """RESTful controller for interactions with categories in the Tool Shed."""

    category_manager: CategoryManager = depends(CategoryManager)

    @expose_api
    @require_admin
    def create(self, trans, payload, **kwd):
        """
        POST /api/categories
        Return a dictionary of information about the created category.
        The following parameters are included in the payload:

        :param name (required): the name of the category
        :param description (optional): the description of the category (if not provided, the name will be used)

        Example: POST /api/categories/?key=XXXYYYXXXYYY
        Content-Disposition: form-data; name="name" Category_Name
        Content-Disposition: form-data; name="description" Category_Description
        """
        category_dict = dict(message="", status="ok")
        request = tool_shed_client.schema.CreateCategoryRequest(
            name=payload.get("name"),
            description=payload.get("description", ""),
        )
        category: Category = self.category_manager.create(trans, request)
        category_dict = self.category_manager.to_dict(category)
        category_dict["message"] = f"Category '{str(category.name)}' has been created"
        return category_dict

    @expose_api_anonymous_and_sessionless
    def get_repositories(self, trans, category_id, **kwd):
        """
        GET /api/categories/{encoded_category_id}/repositories
        Return information about the provided category and the repositories in that category.

        :param id: the encoded id of the Category object
        :param sort_key: the field by which the repositories should be sorted
        :param sort_order: ascending or descending sort
        :param page: the page number to return

        Example: GET localhost:9009/api/categories/f9cad7b01a472135/repositories
        """
        installable = util.asbool(kwd.get("installable", "false"))
        sort_key = kwd.get("sort_key", "name")
        sort_order = kwd.get("sort_order", "asc")
        page = kwd.get("page", None)
        category_dict = repositories_by_category(
            self.app,
            category_id,
            page=page,
            sort_key=sort_key,
            sort_order=sort_order,
            installable=installable,
        )
        category_dict["url"] = web.url_for(controller="categories", action="show", id=category_dict["id"])
        return category_dict

    @expose_api_anonymous_and_sessionless
    def index(self, trans, deleted=False, **kwd) -> List[Dict[str, Any]]:
        """
        GET /api/categories
        Return a list of dictionaries that contain information about each Category.

        :param deleted: flag used to include deleted categories

        Example: GET localhost:9009/api/categories
        """
        deleted = util.asbool(deleted)
        return self.category_manager.index(trans, deleted)

    @expose_api_anonymous_and_sessionless
    def show(self, trans, id, **kwd):
        """
        GET /api/categories/{encoded_category_id}
        Return a dictionary of information about a category.

        :param id: the encoded id of the Category object

        Example: GET localhost:9009/api/categories/f9cad7b01a472135
        """
        category = suc.get_category(self.app, id)
        if category is None:
            category_dict = dict(message=f"Unable to locate category record for id {str(id)}.", status="error")
            return category_dict
        category_dict = self.category_manager.to_dict(category)
        return category_dict



import filecmp
import logging
import os
import tempfile
from typing import TYPE_CHECKING

from galaxy.tool_shed.tools.tool_validator import ToolValidator as GalaxyToolValidator
from galaxy.tools import Tool
from galaxy.util import unicodify
from galaxy.util.tool_shed.xml_util import parse_xml
from tool_shed.util import (
    basic_util,
    hg_util,
    repository_util,
    tool_util,
)

if TYPE_CHECKING:
    from tool_shed.structured_app import ToolShedApp

log = logging.getLogger(__name__)


class ToolValidator(GalaxyToolValidator):
    app: "ToolShedApp"

    def __init__(self, app: "ToolShedApp"):
        super().__init__(app)

    def can_use_tool_config_disk_file(self, repository, repo, file_path, changeset_revision):
        """
        Determine if repository's tool config file on disk can be used.  This method
        is restricted to tool config files since, with the exception of tool config
        files, multiple files with the same name will likely be in various directories
        in the repository and we're comparing file names only (not relative paths).
        """
        if not file_path or not os.path.exists(file_path):
            # The file no longer exists on disk, so it must have been deleted at some previous
            # point in the change log.
            return False
        if changeset_revision == repository.tip():
            return True
        file_name = basic_util.strip_path(file_path)
        latest_version_of_file = self.get_latest_tool_config_revision_from_repository_manifest(
            repo, file_name, changeset_revision
        )
        can_use_disk_file = filecmp.cmp(file_path, latest_version_of_file)
        try:
            os.unlink(latest_version_of_file)
        except Exception:
            pass
        return can_use_disk_file

    def concat_messages(self, msg1, msg2):
        if msg1:
            if msg2:
                message = f"{msg1}  {msg2}"
            else:
                message = msg1
        elif msg2:
            message = msg2
        else:
            message = ""
        return message

    def copy_disk_sample_files_to_dir(self, repo_files_dir, dest_path):
        """
        Copy all files currently on disk that end with the .sample extension to the
        directory to which dest_path refers.
        """
        sample_files = []
        for root, _dirs, files in os.walk(repo_files_dir):
            if root.find(".hg") < 0:
                for name in files:
                    if name.endswith(".sample"):
                        relative_path = os.path.join(root, name)
                        tool_util.copy_sample_file(self.app.config.tool_data_path, relative_path, dest_path=dest_path)
                        sample_files.append(name)
        return sample_files

    def get_latest_tool_config_revision_from_repository_manifest(self, repo, filename, changeset_revision):
        """
        Get the latest revision of a tool config file named filename from the repository
        manifest up to the value of changeset_revision.  This method is restricted to tool_config
        files rather than any file since it is likely that, with the exception of tool config
        files, multiple files will have the same name in various directories within the repository.
        """
        stripped_filename = basic_util.strip_path(filename)
        for changeset in hg_util.reversed_upper_bounded_changelog(repo, changeset_revision):
            manifest_ctx = repo[changeset]
            for ctx_file in manifest_ctx.files():
                ctx_file_name = basic_util.strip_path(unicodify(ctx_file))
                if ctx_file_name == stripped_filename:
                    try:
                        fctx = manifest_ctx[ctx_file]
                    except LookupError:
                        # The ctx_file may have been moved in the change set.  For example,
                        # 'ncbi_blastp_wrapper.xml' was moved to 'tools/ncbi_blast_plus/ncbi_blastp_wrapper.xml',
                        # so keep looking for the file until we find the new location.
                        continue
                    with tempfile.NamedTemporaryFile("wb", prefix="tmp-toolshed-gltcrfrm", delete=False) as fh:
                        tmp_filename = fh.name
                        fh.write(fctx.data())
                    return tmp_filename
        return None

    def get_list_of_copied_sample_files(self, repo, changeset_revision, dir):
        """
        Find all sample files (files in the repository with the special .sample extension)
        in the reversed repository manifest up to changeset_revision. Copy each discovered file to dir and
        return the list of filenames.  If a .sample file was added in a changeset and then
        deleted in a later changeset, it will be returned in the deleted_sample_files list.
        The caller will set the value of app.config.tool_data_path to dir in order to load
        the tools and generate metadata for them.
        """
        deleted_sample_files = []
        sample_files = []
        for changeset in hg_util.reversed_upper_bounded_changelog(repo, changeset_revision):
            changeset_ctx = repo[changeset]
            for ctx_file in changeset_ctx.files():
                ctx_file = unicodify(ctx_file)
                ctx_file_name = basic_util.strip_path(ctx_file)
                # If we decide in the future that files deleted later in the changelog should
                # not be used, we can use the following if statement. if ctx_file_name.endswith( '.sample' )
                # and ctx_file_name not in sample_files and ctx_file_name not in deleted_sample_files:
                if ctx_file_name.endswith(".sample") and ctx_file_name not in sample_files:
                    fctx = hg_util.get_file_context_from_ctx(changeset_ctx, ctx_file)
                    if fctx in ["DELETED"]:
                        # Since the possibly future used if statement above is commented out, the
                        # same file that was initially added will be discovered in an earlier changeset
                        # in the change log and fall through to the else block below.  In other words,
                        # if a file named blast2go.loc.sample was added in change set 0 and then deleted
                        # in changeset 3, the deleted file in changeset 3 will be handled here, but the
                        # later discovered file in changeset 0 will be handled in the else block below.
                        # In this way, the file contents will always be found for future tools even though
                        # the file was deleted.
                        if ctx_file_name not in deleted_sample_files:
                            deleted_sample_files.append(ctx_file_name)
                    else:
                        sample_files.append(ctx_file_name)
                        tmp_ctx_file_name = os.path.join(dir, ctx_file_name.replace(".sample", ""))
                        with open(tmp_ctx_file_name, "wb") as fh:
                            fh.write(fctx.data())
        return sample_files, deleted_sample_files

    def handle_sample_files_and_load_tool_from_disk(
        self, repo_files_dir, repository_id, tool_config_filepath, work_dir
    ):
        """
        Copy all sample files from disk to a temporary directory since the sample files may
        be in multiple directories.
        """
        message = ""
        sample_files = self.copy_disk_sample_files_to_dir(repo_files_dir, work_dir)
        if sample_files:
            if "tool_data_table_conf.xml.sample" in sample_files:
                # Load entries into the tool_data_tables if the tool requires them.
                tool_data_table_config = os.path.join(work_dir, "tool_data_table_conf.xml")
                error, message = self.stdtm.handle_sample_tool_data_table_conf_file(
                    tool_data_table_config, persist=False
                )
        tool, valid, message2 = self.load_tool_from_config(repository_id, tool_config_filepath)
        message = self.concat_messages(message, message2)
        return tool, valid, message, sample_files

    def handle_sample_files_and_load_tool_from_tmp_config(
        self, repo, repository_id, changeset_revision, tool_config_filename, work_dir
    ):
        tool = None
        valid = False
        message = ""
        # We're not currently doing anything with the returned list of deleted_sample_files here.  It is
        # intended to help handle sample files that are in the manifest, but have been deleted from disk.
        sample_files, deleted_sample_files = self.get_list_of_copied_sample_files(
            repo, changeset_revision, dir=work_dir
        )
        if sample_files:
            if "tool_data_table_conf.xml.sample" in sample_files:
                # Load entries into the tool_data_tables if the tool requires them.
                tool_data_table_config = os.path.join(work_dir, "tool_data_table_conf.xml")
                error, message = self.stdtm.handle_sample_tool_data_table_conf_file(
                    tool_data_table_config, persist=False
                )
        manifest_ctx, ctx_file = hg_util.get_ctx_file_path_from_manifest(tool_config_filename, repo, changeset_revision)
        if manifest_ctx and ctx_file:
            tool, valid, message2 = self.load_tool_from_tmp_config(
                repo, repository_id, manifest_ctx, ctx_file, work_dir
            )
            message = self.concat_messages(message, message2)
        return tool, valid, message, sample_files

    def load_tool_from_changeset_revision(self, repository_id, changeset_revision, tool_config_filename):
        """
        Return a loaded tool whose tool config file name (e.g., filtering.xml) is the value
        of tool_config_filename.  The value of changeset_revision is a valid (downloadable)
        changeset revision.  The tool config will be located in the repository manifest between
        the received valid changeset revision and the first changeset revision in the repository,
        searching backwards.
        """
        repository = repository_util.get_repository_in_tool_shed(self.app, repository_id)
        repo_files_dir = repository.repo_path(self.app)
        repo = repository.hg_repo
        tool_config_filepath = repository_util.get_absolute_path_to_file_in_repository(
            repo_files_dir, tool_config_filename
        )
        work_dir = tempfile.mkdtemp(prefix="tmp-toolshed-ltfcr")
        can_use_disk_file = self.can_use_tool_config_disk_file(
            repository, repo, tool_config_filepath, changeset_revision
        )
        if can_use_disk_file:
            tool, valid, message, sample_files = self.handle_sample_files_and_load_tool_from_disk(
                repo_files_dir, repository_id, tool_config_filepath, work_dir
            )
            if tool is not None:
                invalid_files_and_errors_tups = self.check_tool_input_params(
                    repo_files_dir, tool_config_filename, tool, sample_files
                )
                if invalid_files_and_errors_tups:
                    message2 = tool_util.generate_message_for_invalid_tools(
                        self.app,
                        invalid_files_and_errors_tups,
                        repository,
                        metadata_dict=None,
                        as_html=True,
                        displaying_invalid_tool=True,
                    )
                    message = self.concat_messages(message, message2)
        else:
            tool, valid, message, sample_files = self.handle_sample_files_and_load_tool_from_tmp_config(
                repo, repository_id, changeset_revision, tool_config_filename, work_dir
            )
        basic_util.remove_dir(work_dir)
        # Reset the tool_data_tables by loading the empty tool_data_table_conf.xml file.
        self.stdtm.reset_tool_data_tables()
        return repository, tool, valid, message

    def load_tool_from_tmp_config(self, repo, repository_id, ctx, ctx_file, work_dir):
        tool = None
        valid = False
        message = ""
        tmp_tool_config = hg_util.get_named_tmpfile_from_ctx(ctx, ctx_file, work_dir)
        if tmp_tool_config:
            tool_element, error_message = parse_xml(tmp_tool_config)
            if tool_element is None:
                return tool, message
            # Look for external files required by the tool config.
            tmp_code_files = []
            external_paths = Tool.get_externally_referenced_paths(tmp_tool_config)
            changeset_revision = str(ctx)
            for path in external_paths:
                tmp_code_file_name = hg_util.copy_file_from_manifest(repo, changeset_revision, path, work_dir)
                if tmp_code_file_name:
                    tmp_code_files.append(tmp_code_file_name)
            tool, valid, message = self.load_tool_from_config(repository_id, tmp_tool_config)
            for tmp_code_file in tmp_code_files:
                try:
                    os.unlink(tmp_code_file)
                except Exception:
                    pass
            try:
                os.unlink(tmp_tool_config)
            except Exception:
                pass
        return tool, valid, message



import logging
import socket

from markupsafe import escape

from galaxy import (
    util,
    web,
)
from galaxy.managers.api_keys import ApiKeyManager
from galaxy.model.db.user import get_user_by_email
from galaxy.security.validate_user_input import (
    validate_email,
    validate_password,
    validate_publicname,
)
from galaxy.web import url_for
from galaxy.web.form_builder import CheckboxField
from galaxy.webapps.galaxy.controllers.user import User as BaseUser
from tool_shed.webapp import model
from tool_shed.webapp.framework.decorators import require_login

log = logging.getLogger(__name__)

REQUIRE_LOGIN_TEMPLATE = """
<p>
    This %s has been configured such that only users who are logged in may use it.%s
</p>
"""

PASSWORD_RESET_TEMPLATE = """
To reset your Galaxy password for the instance at %s use the following link,
which will expire %s.

%s

If you did not make this request, no action is necessary on your part, though
you may want to notify an administrator.

If you're having trouble using the link when clicking it from email client, you
can also copy and paste it into your browser.
"""


class User(BaseUser):
    @web.expose
    def index(self, trans, cntrller="user", **kwd):
        return trans.fill_template("/webapps/tool_shed/user/index.mako", cntrller=cntrller)

    @web.expose
    def login(self, trans, refresh_frames=None, **kwd):
        """Handle Galaxy Log in"""
        refresh_frames = refresh_frames or []
        referer = trans.request.referer or ""
        redirect = self.__get_redirect_url(kwd.get("redirect", referer).strip())
        redirect_url = ""  # always start with redirect_url being empty
        use_panels = util.string_as_bool(kwd.get("use_panels", False))
        message = kwd.get("message", "")
        status = kwd.get("status", "done")
        header = ""
        user = trans.user
        success = False
        login = kwd.get("login", "")
        if user:
            # Already logged in.
            redirect_url = redirect
            message = "You are already logged in."
            status = "info"
        elif kwd.get("login_button", False):
            response = self.__validate_login(trans, **kwd)
            if trans.response.status == 400:
                trans.response.status = 200
                message = response.get("err_msg")
                status = "error"
            elif response.get("expired_user"):
                change_password_url = url_for(
                    controller="user", action="change_password", id=response.get("expired_user")
                )
                message = f"{response.get('message')}<br>Click <a href='{change_password_url}'>here</a> to change your password."
                status = "warning"
            else:
                success = True
            if success:
                redirect_url = redirect
        if not success and not user and trans.app.config.require_login:
            if trans.app.config.allow_user_creation:
                create_account_str = (
                    "  If you don't already have an account, <a href='{}'>you may create one</a>.".format(
                        web.url_for(controller="user", action="create", cntrller="user")
                    )
                )
                header = REQUIRE_LOGIN_TEMPLATE % ("Galaxy tool shed", create_account_str)
            else:
                header = REQUIRE_LOGIN_TEMPLATE % ("Galaxy tool shed", "")
        return trans.fill_template(
            "/webapps/tool_shed/user/login.mako",
            login=login,
            header=header,
            use_panels=use_panels,
            redirect_url=redirect_url,
            redirect=redirect,
            refresh_frames=refresh_frames,
            message=message,
            status=status,
            form_input_auto_focus=True,
            active_view="user",
        )

    @web.expose
    def create(self, trans, cntrller="user", redirect_url="", refresh_frames=None, **kwd):
        refresh_frames = refresh_frames or []
        params = util.Params(kwd)
        # If the honeypot field is not empty we are dealing with a bot.
        honeypot_field = params.get("bear_field", "")
        if honeypot_field != "":
            return trans.show_error_message(
                'You\'ve been flagged as a possible bot. If you are not, please try registering again and fill the form out carefully. <a target="_top" href="%s">Go to the home page</a>.'
            ) % url_for("/")

        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        use_panels = util.string_as_bool(kwd.get("use_panels", True))
        email = util.restore_text(params.get("email", ""))
        # Do not sanitize passwords, so take from kwd
        # instead of params ( which were sanitized )
        password = kwd.get("password", "")
        confirm = kwd.get("confirm", "")
        username = util.restore_text(params.get("username", ""))
        subscribe = params.get("subscribe", "")
        subscribe_checked = CheckboxField.is_checked(subscribe)
        referer = trans.request.referer or ""
        redirect = kwd.get("redirect", referer).strip()
        is_admin = trans.user_is_admin
        success = False
        show_user_prepopulate_form = False
        if not trans.app.config.allow_user_creation and not trans.user_is_admin:
            message = "User registration is disabled.  Please contact your local Galaxy administrator for an account."
            if trans.app.config.error_email_to is not None:
                message += f" Contact: {trans.app.config.error_email_to}"
            status = "error"
        else:
            # check user is allowed to register
            message, status = trans.app.auth_manager.check_registration_allowed(
                email, username, password, trans.request
            )
            if not message:
                # Create the user, save all the user info and login to Galaxy
                if params.get("create_user_button", False):
                    # Check email and password validity
                    message = self.__validate(trans, email, password, confirm, username)
                    if not message:
                        # All the values are valid
                        message, status, user, success = self.__register(
                            trans, subscribe_checked=subscribe_checked, **kwd
                        )
                        if success and not is_admin:
                            # The handle_user_login() method has a call to the history_set_default_permissions() method
                            # (needed when logging in with a history), user needs to have default permissions set before logging in
                            trans.handle_user_login(user)
                            trans.log_event("User created a new account")
                            trans.log_event("User logged in")
                    else:
                        status = "error"
        registration_warning_message = trans.app.config.registration_warning_message
        if success:
            if is_admin:
                redirect_url = web.url_for("/admin/users?status=success&message=Created new user account.")
            else:
                redirect_url = web.url_for("/")
        return trans.fill_template(
            "/webapps/tool_shed/user/register.mako",
            cntrller=cntrller,
            email=email,
            username=username,
            subscribe_checked=subscribe_checked,
            show_user_prepopulate_form=show_user_prepopulate_form,
            use_panels=use_panels,
            redirect=redirect,
            redirect_url=redirect_url,
            refresh_frames=refresh_frames,
            registration_warning_message=registration_warning_message,
            message=message,
            status=status,
        )

    def __register(self, trans, email=None, username=None, password=None, subscribe_checked=False, **kwd):
        """Registers a new user."""
        email = util.restore_text(email)
        username = util.restore_text(username)
        status = None
        message = None
        is_admin = trans.user_is_admin
        user = self.user_manager.create(email=email, username=username, password=password)
        if subscribe_checked:
            # subscribe user to email list
            if trans.app.config.smtp_server is None:
                status = "error"
                message = f"Now logged in as {user.email}. However, subscribing to the mailing list has failed because mail is not configured for this Galaxy instance. <br>Please contact your local Galaxy administrator."
            else:
                body = "Join Mailing list.\n"
                to = trans.app.config.mailing_join_addr
                frm = email
                subject = "Join Mailing List"
                try:
                    util.send_mail(frm, to, subject, body, trans.app.config)
                except Exception:
                    log.exception("Subscribing to the mailing list has failed.")
                    status = "warning"
                    message = f"Now logged in as {user.email}. However, subscribing to the mailing list has failed."
        if status != "error":
            if not is_admin:
                # The handle_user_login() method has a call to the history_set_default_permissions() method
                # (needed when logging in with a history), user needs to have default permissions set before logging in
                trans.handle_user_login(user)
                trans.log_event("User created a new account")
                trans.log_event("User logged in")
            if trans.app.config.user_activation_on:
                is_activation_sent = self.user_manager.send_activation_email(trans, email, username)
                if is_activation_sent:
                    message = f"Now logged in as {escape(user.email)}.<br>Verification email has been sent to your email address. Please verify it by clicking the activation link in the email.<br>Please check your spam/trash folder in case you cannot find the message.<br><a target=\"_top\" href=\"{url_for('/')}\">Return to the home page.</a>"
                else:
                    status = "error"
                    message = "Unable to send activation email, please contact your local Galaxy administrator."
                    if trans.app.config.error_email_to is not None:
                        message += f" Contact: {trans.app.config.error_email_to}"
        else:
            # User activation is OFF, proceed without sending the activation email.
            message = f"Now logged in as {escape(user.email)}.<br><a target=\"_top\" href=\"{url_for('/')}\">Return to the home page.</a>"
        return message, status, user, status is None

    @web.expose
    def reset_password(self, trans, email=None, **kwd):
        """Reset the user's password. Send an email with token that allows a password change."""
        if trans.app.config.smtp_server is None:
            return trans.show_error_message(
                "Mail is not configured for this Galaxy instance "
                "and password reset information cannot be sent. "
                "Please contact your local Galaxy administrator."
            )
        message = None
        status = "done"
        if kwd.get("reset_password_button", False):
            message = validate_email(trans, email, check_dup=False)
            if not message:
                # Default to a non-userinfo-leaking response message
                message = (
                    f"Your reset request for {escape(email)} has been received.  "
                    "Please check your email account for more instructions.  "
                    "If you do not receive an email shortly, please contact an administrator."
                )
                reset_user = get_user_by_email(trans.sa_session, email, model.User)
                if not reset_user:
                    # Perform a case-insensitive check only if the user wasn't found
                    reset_user = get_user_by_email(trans.sa_session, email, model.User, False)
                if reset_user:
                    prt = model.PasswordResetToken(reset_user)
                    trans.sa_session.add(prt)
                    trans.sa_session.commit()
                    host = trans.request.host.split(":")[0]
                    if host in ["localhost", "127.0.0.1", "0.0.0.0"]:
                        host = socket.getfqdn()
                    reset_url = url_for(controller="user", action="change_password", token=prt.token, qualified=True)
                    body = PASSWORD_RESET_TEMPLATE % (
                        host,
                        prt.expiration_time.strftime(trans.app.config.pretty_datetime_format),
                        reset_url,
                    )
                    frm = trans.app.config.email_from or f"galaxy-no-reply@{host}"
                    subject = "Galaxy Password Reset"
                    try:
                        util.send_mail(frm, email, subject, body, trans.app.config)
                        trans.sa_session.add(reset_user)
                        trans.sa_session.commit()
                        trans.log_event(f"User reset password: {email}")
                    except Exception:
                        log.exception("Unable to reset password.")
        return trans.fill_template("/webapps/tool_shed/user/reset_password.mako", message=message, status=status)

    @web.expose
    def manage_user_info(self, trans, cntrller, **kwd):
        """Manage a user's login, password, public username, type, addresses, etc."""
        params = util.Params(kwd)
        user_id = params.get("id", None)
        if user_id:
            user = trans.sa_session.get(model.User, trans.security.decode_id(user_id))
        else:
            user = trans.user
        if not user:
            raise AssertionError(f"The user id ({str(user_id)}) is not valid")
        email = util.restore_text(params.get("email", user.email))
        username = util.restore_text(params.get("username", ""))
        if not username:
            username = user.username
        message = escape(util.restore_text(params.get("message", "")))
        status = params.get("status", "done")
        return trans.fill_template(
            "/webapps/tool_shed/user/manage_info.mako",
            cntrller=cntrller,
            user=user,
            email=email,
            username=username,
            message=message,
            status=status,
        )

    @web.expose
    @require_login()
    def api_keys(self, trans, cntrller, **kwd):
        params = util.Params(kwd)
        message = escape(util.restore_text(params.get("message", "")))
        status = params.get("status", "done")
        if params.get("new_api_key_button", False):
            ApiKeyManager(trans.app).create_api_key(trans.user)
            message = "Generated a new web API key"
            status = "done"
        return trans.fill_template(
            "/webapps/tool_shed/user/api_keys.mako", cntrller=cntrller, user=trans.user, message=message, status=status
        )

    # For REMOTE_USER, we need the ability to just edit the username
    @web.expose
    @require_login("to manage the public name")
    def edit_username(self, trans, cntrller, **kwd):
        params = util.Params(kwd)
        is_admin = cntrller == "admin" and trans.user_is_admin
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        user_id = params.get("user_id", None)
        if user_id and is_admin:
            user = trans.sa_session.get(model.User, trans.security.decode_id(user_id))
        else:
            user = trans.user
        if user and params.get("change_username_button", False):
            username = kwd.get("username", "")
            if username:
                message = validate_publicname(trans, username, user)
            if message:
                status = "error"
            else:
                user.username = username
                trans.sa_session.add(user)
                trans.sa_session.commit()
                message = "The username has been updated with the changes."
        return trans.fill_template(
            "/webapps/tool_shed/user/username.mako",
            cntrller=cntrller,
            user=user,
            username=user.username,
            message=message,
            status=status,
        )

    @web.expose
    def edit_info(self, trans, cntrller, **kwd):
        """
        Edit user information = username, email or password.
        """
        params = util.Params(kwd)
        is_admin = cntrller == "admin" and trans.user_is_admin
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        user_id = params.get("user_id", None)
        if user_id and is_admin:
            user = trans.sa_session.get(model.User, trans.security.decode_id(user_id))
        elif user_id and (not trans.user or trans.user.id != trans.security.decode_id(user_id)):
            message = "Invalid user id"
            status = "error"
            user = None
        else:
            user = trans.user
        if user and params.get("login_info_button", False):
            # Editing email and username
            email = util.restore_text(params.get("email", ""))
            username = util.restore_text(params.get("username", "")).lower()

            # Validate the new values for email and username
            message = validate_email(trans, email, user)
            if not message and username:
                message = validate_publicname(trans, username, user)
            if message:
                status = "error"
            else:
                if user.email != email:
                    # The user's private role name must match the user's login ( email )
                    private_role = trans.app.security_agent.get_private_user_role(user)
                    private_role.name = email
                    private_role.description = f"Private role for {email}"
                    # Change the email itself
                    user.email = email
                    trans.sa_session.add_all((user, private_role))
                    trans.sa_session.commit()
                    if trans.webapp.name == "galaxy" and trans.app.config.user_activation_on:
                        user.active = False
                        trans.sa_session.add(user)
                        trans.sa_session.commit()
                        is_activation_sent = self.user_manager.send_activation_email(trans, user.email, user.username)
                        if is_activation_sent:
                            message = "The login information has been updated with the changes.<br>Verification email has been sent to your new email address. Please verify it by clicking the activation link in the email.<br>Please check your spam/trash folder in case you cannot find the message."
                        else:
                            message = "Unable to send activation email, please contact your local Galaxy administrator."
                            if trans.app.config.error_email_to is not None:
                                message += f" Contact: {trans.app.config.error_email_to}"
                if user.username != username:
                    user.username = username
                    trans.sa_session.add(user)
                    trans.sa_session.commit()
                message = "The login information has been updated with the changes."
        if user and trans.webapp.name == "galaxy" and is_admin:
            kwd["user_id"] = trans.security.encode_id(user.id)
        kwd["id"] = user_id
        if message:
            kwd["message"] = util.sanitize_text(message)
        if status:
            kwd["status"] = status
        return trans.response.send_redirect(
            web.url_for(controller="user", action="manage_user_info", cntrller=cntrller, **kwd)
        )

    @web.expose
    def change_password(self, trans, token=None, id=None, **kwd):
        """
        Provides a form with which one can change their password.  If token is
        provided, don't require current password.
        """
        if kwd.get("change_password_button", False):
            password = kwd.get("password", "")
            confirm = kwd.get("confirm", "")
            current = kwd.get("current", "")
            user, message = self.user_manager.change_password(
                trans, password=password, current=current, token=token, confirm=confirm, id=id
            )
            if not user:
                return trans.show_error_message(message)
            return trans.show_ok_message(
                "The password has been changed and any other existing Galaxy sessions have been logged out (but jobs in histories in those sessions will not be interrupted)."
            )
        return trans.fill_template("/webapps/tool_shed/user/change_password.mako", token=token, id=id)

    @web.expose
    def logout(self, trans, logout_all=False, **kwd):
        trans.handle_user_logout(logout_all=logout_all)
        message = f'You have been logged out.<br>To log in again <a target="_top" href="{url_for("/")}">go to the home page</a>.'
        if trans.app.config.use_remote_user and trans.app.config.remote_user_logout_href:
            trans.response.send_redirect(trans.app.config.remote_user_logout_href)
        else:
            return trans.fill_template(
                "/webapps/tool_shed/user/logout.mako",
                refresh_frames=["masthead"],
                message=message,
                status="done",
                active_view="user",
            )

    def __validate(self, trans, email, password, confirm, username):
        if username in ["repos"]:
            return f"The term '{username}' is a reserved word in the Tool Shed, so it cannot be used as a public user name."
        message = "\n".join(
            (
                validate_email(trans, email),
                validate_password(trans, password, confirm),
                validate_publicname(trans, username),
            )
        ).rstrip()
        return message



import logging
from json import JSONDecodeError
from typing import (
    AsyncGenerator,
    cast,
    List,
    Optional,
    Type,
    TypeVar,
)

from fastapi import (
    Depends,
    HTTPException,
    Path,
    Query,
    Request,
    Response,
    Security,
)
from fastapi.security import (
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
)
from pydantic import BaseModel
from starlette_context import context as request_context

from galaxy.exceptions import AdminRequiredException
from galaxy.managers.session import GalaxySessionManager
from galaxy.managers.users import UserManager
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.util import unicodify
from galaxy.web.framework.decorators import require_admin_message
from galaxy.webapps.base.webapp import create_new_session
from galaxy.webapps.galaxy.api import (
    depends as framework_depends,
    FrameworkRouter,
    GalaxyASGIRequest,
    GalaxyASGIResponse,
    T,
    UrlBuilder,
)
from tool_shed.context import (
    SessionRequestContext,
    SessionRequestContextImpl,
)
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp import app as tool_shed_app_mod
from tool_shed.webapp.model import (
    GalaxySession,
    User,
)
from tool_shed_client.schema import IndexSortByType

log = logging.getLogger(__name__)


def get_app() -> ToolShedApp:
    if tool_shed_app_mod.app is None:
        raise Exception("Failed to initialize the tool shed app correctly for FastAPI")
    return tool_shed_app_mod.app


async def get_app_with_request_session() -> AsyncGenerator[ToolShedApp, None]:
    app = get_app()
    request_id = request_context.data["X-Request-ID"]
    app.model.set_request_id(request_id)
    try:
        yield app
    finally:
        app.model.unset_request_id(request_id)


DependsOnApp = cast(ToolShedApp, Depends(get_app_with_request_session))
AUTH_COOKIE_NAME = "galaxycommunitysession"

api_key_query = APIKeyQuery(name="key", auto_error=False)
api_key_header = APIKeyHeader(name="x-api-key", auto_error=False)
api_key_cookie = APIKeyCookie(name=AUTH_COOKIE_NAME, auto_error=False)


def depends(dep_type: Type[T]) -> T:
    return framework_depends(dep_type, app=get_app_with_request_session)


def get_api_user(
    user_manager: UserManager = depends(UserManager),
    key: str = Security(api_key_query),
    x_api_key: str = Security(api_key_header),
) -> Optional[User]:
    api_key = key or x_api_key
    if not api_key:
        return None
    user = user_manager.by_api_key(api_key=api_key)
    return user


def get_session_manager(app: ToolShedApp = DependsOnApp) -> GalaxySessionManager:
    # TODO: find out how to adapt dependency for Galaxy/Report/TS
    return GalaxySessionManager(app.model)


def get_session(
    session_manager=cast(GalaxySessionManager, Depends(get_session_manager)),
    security: IdEncodingHelper = depends(IdEncodingHelper),
    galaxysession: str = Security(api_key_cookie),
) -> Optional[GalaxySession]:
    if galaxysession:
        session_key = security.decode_guid(galaxysession)
        if session_key:
            return session_manager.get_session_from_session_key(session_key)
        # TODO: What should we do if there is no session? Since this is the API, maybe nothing is the right choice?
    return None


def get_user(
    galaxy_session=cast(Optional[GalaxySession], Depends(get_session)),
    api_user=cast(Optional[User], Depends(get_api_user)),
) -> Optional[User]:
    if galaxy_session:
        return galaxy_session.user
    return api_user


def get_trans(
    request: Request,
    response: Response,
    app: ToolShedApp = DependsOnApp,
    user=cast(Optional[User], Depends(get_user)),
    galaxy_session=cast(Optional[GalaxySession], Depends(get_session)),
) -> SessionRequestContext:
    url_builder = UrlBuilder(request)
    galaxy_request = GalaxyASGIRequest(request)
    galaxy_response = GalaxyASGIResponse(response)
    return SessionRequestContextImpl(
        app,
        galaxy_request,
        galaxy_response,
        user=user,
        galaxy_session=galaxy_session,
        url_builder=url_builder,
    )


DependsOnTrans: SessionRequestContext = cast(SessionRequestContext, Depends(get_trans))


def get_admin_user(trans: SessionRequestContext = DependsOnTrans):
    if not trans.user_is_admin:
        raise AdminRequiredException(require_admin_message(trans.app.config, trans.user))
    return trans.user


AdminUserRequired = Depends(get_admin_user)


class Router(FrameworkRouter):
    admin_user_dependency = AdminUserRequired


B = TypeVar("B", bound=BaseModel)


# async def depend_on_either_json_or_form_data(model: Type[T]):
#    return Depends(get_body)


def depend_on_either_json_or_form_data(model: Type[B]) -> B:
    async def get_body(request: Request):
        content_type = request.headers.get("Content-Type")
        if content_type is None:
            raise HTTPException(status_code=400, detail="No Content-Type provided!")
        elif content_type == "application/json":
            try:
                return model(**await request.json())
            except JSONDecodeError:
                raise HTTPException(status_code=400, detail="Invalid JSON data")
        elif content_type == "application/x-www-form-urlencoded" or content_type.startswith("multipart/form-data"):
            try:
                return model(**await request.form())
            except Exception:
                raise HTTPException(status_code=400, detail="Invalid Form data")
        else:
            raise HTTPException(status_code=400, detail="Content-Type not supported!")

    return Depends(get_body)


UserIdPathParam: str = Path(..., title="User ID", description="The encoded database identifier of the user.")

RequiredRepoOwnerParam: str = Query(
    title="owner",
    description="Owner of the target repository.",
)

RequiredRepoNameParam: str = Query(
    title="Name",
    description="Name of the target repository.",
)

RequiredChangesetParam: str = Query(
    title="changeset",
    description="Changeset of the target repository.",
)

RepositoryIdPathParam: str = Path(
    ..., title="Repository ID", description="The encoded database identifier of the repository."
)

ChangesetRevisionPathParam: str = Path(
    ...,
    title="Change Revision",
    description="The changeset revision corresponding to the target revision of the target repository.",
)

UsernameIdPathParam: str = Path(..., title="Username", description="The target username.")

CommitMessageQueryParam: Optional[str] = Query(
    default=None,
    title="Commit Message",
    description="Set commit message as a query parameter.",
)

DownloadableQueryParam: bool = Query(
    default=True,
    title="downloadable_only",
    description="Include only downloadable repositories.",
)

CommitMessage: str = Query(
    None,
    title="Commit message",
    description="A commit message to store with repository update.",
)

RepositoryIndexQueryParam: Optional[str] = Query(
    default=None,
    title="Search Query",
    description="This will perform a full search with whoosh on the backend and will cause the API endpoint to return a RepositorySearchResult. This should not be used with the 'filter' parameter.",
)

RepositoryIndexFilterParam: Optional[str] = Query(
    default=None,
    title="Filter Text",
    description="This will perform a quick search using database operators. This should not be used with the 'q' parameter.",
)

RepositoryIndexSortByParam: IndexSortByType = Query(
    default="name",
    title="Sort by",
    description="Sort by the this repository field - direction is controlled by sort_desc that defaults to False and causes an ascending sort on this field. This field is ignored if 'q' is specified an whoosh search is used.",
)

RepositoryIndexSortDescParam: bool = Query(
    default=False,
    title="Sort Descending",
    description="Direction of sort. This defaults to False and causes an ascending sort on the field specified by sort_on. This field is ignored if 'q' is specified an whoosh search is used.",
)


ToolsIndexQueryParam: str = Query(
    default=...,
    title="Search Query",
)

ToolSearchPageQueryParam: int = Query(
    default=1,
    title="Page",
    description="",
)

RepositorySearchPageQueryParam: Optional[int] = Query(
    default=None,
    title="Page",
    description="",
)

RepositorySearchPageSizeQueryParam: int = Query(
    default=10,
    title="Page Size",
)

RepositoryIndexDeletedQueryParam: Optional[bool] = Query(False, title="Deleted?")

RepositoryIndexOwnerQueryParam: Optional[str] = Query(None, title="Owner")

RepositoryIndexNameQueryParam: Optional[str] = Query(None, title="Name")

RepositoryIndexCategoryQueryParam: Optional[str] = Query(None, title="Category ID")

RepositoryIndexToolIdsQueryParam: Optional[List[str]] = Query(
    None, title="Tool IDs", description="List of tool GUIDs to find the repository for"
)


OptionalRepositoryOwnerParam: Optional[str] = Query(None, title="Owner")
OptionalRepositoryNameParam: Optional[str] = Query(None, title="Name")
RequiredRepositoryChangesetRevisionParam: str = Query(..., title="Changeset Revision")
OptionalRepositoryIdParam: Optional[str] = Query(None, title="TSR ID")
OptionalHexlifyParam: Optional[bool] = Query(True, title="Hexlify response")

CategoryIdPathParam: str = Path(
    ..., title="Category ID", description="The encoded database identifier of the category."
)
CategoryRepositoriesInstallableQueryParam: bool = Query(False, title="Installable?")
CategoryRepositoriesSortKeyQueryParam: str = Query("name", title="Sort Key")
CategoryRepositoriesSortOrderQueryParam: str = Query("asc", title="Sort Order")
CategoryRepositoriesPageQueryParam: Optional[int] = Query(None, title="Page")


def ensure_valid_session(trans: SessionRequestContext) -> None:
    """
    Ensure that a valid Galaxy session exists and is available as
    trans.session (part of initialization)
    """
    app = trans.app
    mapping = app.model
    session_manager = GalaxySessionManager(mapping)
    sa_session = app.model.context
    request = trans.request
    # Try to load an existing session
    galaxy_session = None
    prev_galaxy_session = None
    user_for_new_session = None
    invalidate_existing_session = False
    # Track whether the session has changed so we can avoid calling flush
    # in the most common case (session exists and is valid).
    galaxy_session_requires_flush = False
    if secure_id := request.get_cookie(AUTH_COOKIE_NAME):
        session_key: Optional[str] = app.security.decode_guid(secure_id)
        if session_key:
            # We do NOT catch exceptions here, if the database is down the request should fail,
            # and we should not generate a new session.
            galaxy_session = session_manager.get_session_from_session_key(session_key=session_key)
        if not galaxy_session:
            session_key = None

    if galaxy_session is not None and galaxy_session.user is not None and galaxy_session.user.deleted:
        invalidate_existing_session = True
        log.warning(f"User '{galaxy_session.user.email}' is marked deleted, invalidating session")
    # Do we need to invalidate the session for some reason?
    if invalidate_existing_session:
        assert galaxy_session
        prev_galaxy_session = galaxy_session
        prev_galaxy_session.is_valid = False
        galaxy_session = None
    # No relevant cookies, or couldn't find, or invalid, so create a new session
    if galaxy_session is None:
        galaxy_session = create_new_session(trans, prev_galaxy_session, user_for_new_session)
        galaxy_session_requires_flush = True
        trans.set_galaxy_session(galaxy_session)
        set_auth_cookie(trans, galaxy_session)
    else:
        trans.set_galaxy_session(galaxy_session)
    # Do we need to flush the session?
    if galaxy_session_requires_flush:
        sa_session.add(galaxy_session)
        # FIXME: If prev_session is a proper relation this would not
        #        be needed.
        if prev_galaxy_session:
            sa_session.add(prev_galaxy_session)
        sa_session.commit()


def set_auth_cookie(trans: SessionRequestContext, session):
    cookie_name = AUTH_COOKIE_NAME
    set_cookie(trans, trans.app.security.encode_guid(session.session_key), cookie_name)


def set_cookie(trans: SessionRequestContext, value: str, key, path="/", age=90) -> None:
    """Convenience method for setting a session cookie"""
    # In wsgi we were setting both a max_age and and expires, but
    # all browsers support max_age now.
    domain: Optional[str] = trans.app.config.cookie_domain
    trans.response.set_cookie(
        key,
        unicodify(value),
        path=path,
        max_age=3600 * 24 * age,  # 90 days
        httponly=True,
        secure=trans.request.is_secure,
        domain=domain,
    )



from ..base.api import ShedApiTestCase


class TestShedGalaxyInstallApi(ShedApiTestCase):
    def test_install_simple_tool(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoformetadata")
        owner = repository.owner
        name = repository.name
        installable_revisions = populator.get_ordered_installable_revisions(owner, name)
        latest_install_revision = installable_revisions.root[-1]
        self.install_repository(owner, name, latest_install_revision, tool_shed_url=self.url)
        response = self.galaxy_interactor._get("tools?in_panel=False")
        response.raise_for_status()
        expected_tool = populator.tool_guid(self, repository, "Add_a_column1", "1.1.0")
        tool_ids = [t["id"] for t in response.json()]
        assert expected_tool in tool_ids, f"Didn't find {expected_tool} in {tool_ids}"

    def test_install_simple_after_repository_metadata_reset(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoformetadata")
        owner = repository.owner
        name = repository.name
        installable_revisions = populator.get_ordered_installable_revisions(owner, name)
        latest_install_revision = installable_revisions.root[-1]
        metadata_response = populator.reset_metadata(repository)
        assert metadata_response.status == "ok"
        self.install_repository(owner, name, latest_install_revision, tool_shed_url=self.url)
        response = self.galaxy_interactor._get("tools?in_panel=False")
        response.raise_for_status()
        expected_tool = f"{self.host}:{self.port}/repos/{owner}/{name}/Add_a_column1/1.1.0"
        tool_ids = [t["id"] for t in response.json()]
        assert expected_tool in tool_ids, f"Didn't find {expected_tool} in {tool_ids}"



from playwright.sync_api import (
    expect,
    Page,
)

from galaxy_test.base.api_util import random_name
from ..base.api import skip_if_api_v1
from ..base.playwrightbrowser import (
    Locators,
    PlaywrightShedBrowser,
)
from ..base.twilltestcase import ShedTwillTestCase


class PlaywrightTestCase(ShedTwillTestCase):
    @property
    def _playwright_browser(self) -> PlaywrightShedBrowser:
        browser = self._browser
        assert isinstance(browser, PlaywrightShedBrowser)
        return browser

    @property
    def _page(self) -> Page:
        return self._playwright_browser._page


TEST_PASSWORD = "testpass"


class TestFrontendLogin(PlaywrightTestCase):
    @skip_if_api_v1
    def test_register(self):
        self.visit_url("/")
        page = self._page
        expect(page.locator(Locators.toolbar_login)).to_be_visible()
        page.click(Locators.toolbar_login)
        expect(page.locator(Locators.login_submit_button)).to_be_visible()
        expect(page.locator(Locators.register_link)).to_be_visible()
        page.click(Locators.register_link)
        user = random_name(prefix="shduser")
        self._submit_register_form(
            f"{user}@galaxyproject.org",
            TEST_PASSWORD,
            user,
        )
        expect(page.locator(Locators.login_submit_button)).to_be_visible()

    @skip_if_api_v1
    def test_create(self):
        user = random_name(prefix="shduser")
        self.create(
            email=f"{user}@galaxyproject.org",
            password=TEST_PASSWORD,
            username=user,
        )

    @skip_if_api_v1
    def test_logout(self):
        self._create_and_login()
        self._playwright_browser.expect_logged_in()
        self._playwright_browser.logout_if_logged_in()
        self._playwright_browser.expect_not_logged_in()

    @skip_if_api_v1
    def test_change_password(self):
        self._create_and_login()

    def _create_and_login(self):
        user = random_name(prefix="shduser")
        email = f"{user}@galaxyproject.org"
        self.create(
            email=email,
            password=TEST_PASSWORD,
            username=user,
        )
        self.login(email, TEST_PASSWORD, username=user)



"""Module for searching the toolshed tools within all repositories"""

import logging
import os

import whoosh.index
from whoosh import scoring
from whoosh.fields import (
    ID,
    Schema,
    TEXT,
)
from whoosh.qparser import MultifieldParser

from galaxy import exceptions
from galaxy.exceptions import ObjectNotFound
from galaxy.util import unicodify

log = logging.getLogger(__name__)

schema = Schema(
    name=TEXT(stored=True),
    description=TEXT(stored=True),
    owner=TEXT(stored=True),
    id=TEXT(stored=True),
    help=TEXT(stored=True),
    version=TEXT(stored=True),
    repo_name=TEXT(stored=True),
    repo_owner_username=TEXT(stored=True),
    repo_id=ID(stored=True),
)


class ToolSearch:
    def search(self, app, search_term, page, page_size, boosts):
        """
        Perform the search on the given search_term

        :param search_term: unicode encoded string with the search term(s)

        :returns results: dictionary containing number of hits, hits themselves and matched terms for each
        """
        tool_index_dir = os.path.join(app.config.whoosh_index_dir, "tools")
        index_exists = whoosh.index.exists_in(tool_index_dir)
        if index_exists:
            index = whoosh.index.open_dir(tool_index_dir)
            try:
                # Some literature about BM25F:
                # http://trec.nist.gov/pubs/trec13/papers/microsoft-cambridge.web.hard.pdf
                # http://en.wikipedia.org/wiki/Okapi_BM25
                # __Basically__ the higher number the bigger weight.
                tool_weighting = scoring.BM25F(
                    field_B={
                        "name_B": boosts.tool_name_boost,
                        "description_B": boosts.tool_description_boost,
                        "help_B": boosts.tool_help_boost,
                        "repo_owner_username_B": boosts.tool_repo_owner_username_boost,
                    }
                )
                searcher = index.searcher(weighting=tool_weighting)

                parser = MultifieldParser(["name", "description", "help", "repo_owner_username"], schema=schema)

                user_query = parser.parse(f"*{search_term}*")

                try:
                    hits = searcher.search_page(user_query, page, pagelen=page_size, terms=True)
                except ValueError:
                    raise ObjectNotFound("The requested page does not exist.")

                log.debug(f"searching tools for: #{str(search_term)}")
                log.debug(f"total hits: {str(len(hits))}")
                log.debug(f"scored hits: {str(hits.scored_length())}")
                results = {}
                results["total_results"] = str(len(hits))
                results["page"] = str(page)
                results["page_size"] = str(page_size)
                results["hits"] = []
                for hit in hits:
                    hit_dict = {}
                    hit_dict["id"] = hit.get("id")
                    hit_dict["repo_owner_username"] = hit.get("repo_owner_username")
                    hit_dict["repo_name"] = hit.get("repo_name")
                    hit_dict["name"] = hit.get("name")
                    hit_dict["description"] = hit.get("description")
                    matched_terms = {k: unicodify(v) for k, v in hit.matched_terms()}
                    results["hits"].append({"tool": hit_dict, "matched_terms": matched_terms, "score": hit.score})
                return results
            finally:
                searcher.close()
        else:
            raise exceptions.InternalServerError("The search index file is missing.")



import os
from typing import (
    Callable,
    Generator,
)

import pytest
from playwright.sync_api import Browser

from ..base.browser import ShedBrowser
from ..base.playwrightbrowser import PlaywrightShedBrowser
from ..base.twillbrowser import TwillShedBrowser

DEFAULT_BROWSER = "playwright"


def twill_browser() -> Generator[ShedBrowser, None, None]:
    yield TwillShedBrowser()


def playwright_browser(browser: Browser) -> Generator[ShedBrowser, None, None]:
    page = browser.new_page()
    yield PlaywrightShedBrowser(page)


test_browser = os.environ.get("TOOL_SHED_TEST_BROWSER", DEFAULT_BROWSER)
if test_browser == "twill":
    shed_browser: Callable[..., Generator[ShedBrowser, None, None]] = pytest.fixture(scope="class")(twill_browser)
elif test_browser == "playwright":
    shed_browser = pytest.fixture(scope="class")(playwright_browser)
else:
    raise ValueError(f"Unrecognized value for TOOL_SHED_TEST_BROWSER: {test_browser}")



import os
import tempfile

from galaxy.tool_util.parser import get_tool_source
from galaxy.util.compression_utils import CompressedFile
from galaxy.util.resources import resource_path
from galaxy_test.base import api_asserts
from tool_shed.test.base.api_util import create_user
from tool_shed.test.base.populators import (
    HasRepositoryId,
    repo_tars,
)
from tool_shed_client.schema import (
    RepositoryIndexRequest,
    RepositoryPaginatedIndexRequest,
    RepositoryRevisionMetadata,
    UpdateRepositoryRequest,
)
from ..base.api import (
    ShedApiTestCase,
    skip_if_api_v1,
    skip_if_api_v2,
)

COLUMN_MAKER_PATH = resource_path(__name__, "../test_data/column_maker/column_maker.tar")


# test_0000 tests commit_message  - find a way to test it here
class TestShedRepositoriesApi(ShedApiTestCase):
    def test_create(self):
        populator = self.populator
        category1_id = populator.new_category(prefix="testcreate").id
        populator.assert_category_has_n_repositories(category1_id, 0)

        populator.new_repository(category1_id)
        populator.assert_category_has_n_repositories(category1_id, 1)

        # Test creating repository with multiple categories
        category2_id = populator.new_category(prefix="testcreate").id
        populator.assert_category_has_n_repositories(category2_id, 0)
        populator.new_repository([category1_id, category2_id])
        populator.assert_category_has_n_repositories(category1_id, 2)
        populator.assert_category_has_n_repositories(category2_id, 1)

    def test_update_repository(self):
        populator = self.populator
        prefix = "testupdate"
        category_id = populator.new_category(prefix=prefix).id
        repository = populator.new_repository(category_id, prefix=prefix)
        repository_id = repository.id
        repository_update = populator.upload_revision(
            repository_id,
            COLUMN_MAKER_PATH,
        )
        assert repository_update.is_ok

    def test_update_repository_info(self):
        populator = self.populator
        prefix = "testupdateinfo"
        category_id = populator.new_category(prefix=prefix).id
        repository = populator.new_repository(category_id, prefix=prefix)
        repository_id = repository.id
        request = UpdateRepositoryRequest(homepage_url="https://www.google.com")
        update = populator.update(repository_id, request)
        assert update.homepage_url == "https://www.google.com"
        assert populator.get_repository(repository_id).homepage_url == "https://www.google.com"

    def test_update_category(self):
        populator = self.populator
        prefix = "testupdatecategory"

        old_category_id = populator.new_category(prefix=prefix).id
        new_category_id = populator.new_category(prefix=prefix).id

        populator.assert_category_has_n_repositories(old_category_id, 0)
        populator.assert_category_has_n_repositories(new_category_id, 0)

        repository = populator.new_repository(old_category_id, prefix=prefix)
        repository_id = repository.id

        populator.assert_category_has_n_repositories(old_category_id, 1)
        populator.assert_category_has_n_repositories(new_category_id, 0)

        request = UpdateRepositoryRequest(category_ids=[new_category_id])
        populator.update(repository_id, request)

        # does this mean we cannot remove categories?
        populator.assert_category_has_n_repositories(old_category_id, 0)
        populator.assert_category_has_n_repositories(new_category_id, 1)

    # used by getRepository in TS client.
    def test_metadata_simple(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoformetadata")
        repository_metadata = populator.get_metadata(repository)
        metadata_for_revisions = repository_metadata.root
        assert len(metadata_for_revisions) == 1
        only_key = list(metadata_for_revisions.keys())[0]
        assert only_key.startswith("0:")
        only_revision = list(metadata_for_revisions.values())[0]
        assert only_revision
        assert only_revision.downloadable
        assert not only_revision.malicious

    def test_metadata_invalid_tools(self):
        populator = self.populator
        repository = populator.setup_bismark_repo()
        repository_metadata = populator.get_metadata(repository)
        assert repository_metadata
        for _, value in repository_metadata.root.items():
            assert value.invalid_tools

    def test_index_simple(self):
        # Logic and typing is pretty different if given a tool id to search for - this should
        # be tested or dropped in v2.
        populator = self.populator
        repo = populator.setup_column_maker_repo(prefix="repoforindex")
        repository_id = repo.id
        show_response = self.api_interactor.get(f"repositories/{repository_id}")
        index_response = self.api_interactor.get("repositories")
        api_asserts.assert_status_code_is_ok(show_response)
        api_asserts.assert_status_code_is_ok(index_response)
        repository_ids = [r["id"] for r in index_response.json()]
        assert repository_id in repository_ids

        repository = self.populator.get_repository_for(repo.owner, repo.name)
        assert repository
        assert repository.owner == repo.owner
        assert repository.name == repo.name

    @skip_if_api_v1
    def test_index_pagination(self):
        populator = self.populator
        category1 = populator.new_category(prefix="paginatecat1")
        category2 = populator.new_category(prefix="paginatecat2")
        populator.setup_column_maker_repo(prefix="repoforindexpagination1", category_id=category1.id)
        populator.setup_column_maker_repo(prefix="repoforindexpagination2", category_id=category1.id)
        populator.setup_column_maker_repo(prefix="repoforindexpagination3", category_id=category2.id)
        populator.setup_column_maker_repo(prefix="repoforindexpagination4", category_id=category2.id)
        populator.setup_column_maker_repo(prefix="repoforindexpagination5", category_id=category2.id)
        request = RepositoryPaginatedIndexRequest(
            page=1,
            page_size=2,
            category_id=category1.id,
        )
        response = populator.repository_index_paginated(request)
        assert len(response.hits) == 2
        assert response.total_results == 2
        assert response.page == 1
        assert response.page_size == 2

        request.category_id = category2.id
        response = populator.repository_index_paginated(request)
        assert response.total_results == 3
        assert response.page == 1
        assert response.page_size == 2

        request.filter = "repoforindexpagination4"
        response = populator.repository_index_paginated(request)
        assert response.total_results == 1

    @skip_if_api_v1
    def test_index_sorting(self):
        populator = self.populator
        category1 = populator.new_category(prefix="paginatecat1")
        populator.setup_column_maker_repo(prefix="repoforsort_z", category_id=category1.id)
        populator.setup_column_maker_repo(prefix="repoforsort_a", category_id=category1.id)

        response = populator.repository_index(RepositoryIndexRequest())
        order_of_these = [r.name for r in response.root if r.name.startswith("repoforsort")]
        assert "_a" in order_of_these[0]
        assert "_z" in order_of_these[1]

        response = populator.repository_index(RepositoryIndexRequest(sort_desc=True))
        order_of_these = [r.name for r in response.root if r.name.startswith("repoforsort")]
        assert "_z" in order_of_these[0]
        assert "_a" in order_of_these[1]

        # test recently created query
        response = populator.repository_index(RepositoryIndexRequest(sort_desc=True, sort_by="create_time"))
        order_of_these = [r.name for r in response.root if r.name.startswith("repoforsort")]
        assert "_a" in order_of_these[0]
        assert "_z" in order_of_these[1]

    @skip_if_api_v1
    def test_allow_push(self):
        populator = self.populator
        request = {
            "email": "sharewith@galaxyproject.org",
            "username": "sharewith",
            "password": "pAssworD1",
        }
        create_user(self.admin_api_interactor, request)
        request = {
            "email": "alsosharewith@galaxyproject.org",
            "username": "alsosharewith",
            "password": "pAssworD2",
        }
        create_user(self.admin_api_interactor, request)

        repo = populator.setup_column_maker_repo(prefix="repoforindex")
        assert "sharewith" not in populator.get_usernames_allowed_to_push(repo)
        assert "alsosharewith" not in populator.get_usernames_allowed_to_push(repo)

        populator.allow_user_to_push(repo, "sharewith")
        assert "sharewith" in populator.get_usernames_allowed_to_push(repo)
        assert "alsosharewith" not in populator.get_usernames_allowed_to_push(repo)

        populator.allow_user_to_push(repo, "alsosharewith")
        assert "sharewith" in populator.get_usernames_allowed_to_push(repo)
        assert "alsosharewith" in populator.get_usernames_allowed_to_push(repo)

        populator.disallow_user_to_push(repo, "sharewith")
        assert "sharewith" not in populator.get_usernames_allowed_to_push(repo)
        assert "alsosharewith" in populator.get_usernames_allowed_to_push(repo)

    @skip_if_api_v1
    def test_set_malicious(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoformalicious")

        only_revision = self._get_only_revision(repository)
        assert only_revision.downloadable
        assert not only_revision.malicious

        assert not populator.tip_is_malicious(repository)
        populator.set_malicious(repository, only_revision.changeset_revision)
        assert populator.tip_is_malicious(repository)
        populator.unset_malicious(repository, only_revision.changeset_revision)
        assert not populator.tip_is_malicious(repository)

    @skip_if_api_v1
    def test_set_deprecated(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repofordeprecated")
        assert not repository.deprecated
        assert not populator.is_deprecated(repository)
        populator.set_deprecated(repository)
        assert populator.is_deprecated(repository)
        populator.unset_deprecated(repository)
        assert not populator.is_deprecated(repository)

    def test_install_info(self):
        # actually installing requires a whole Galaxy setup and the install manager but
        # we can test the response validates against the future facing InstallInfo pydandic
        # models.
        populator = self.populator
        repo = populator.setup_column_maker_and_get_metadata(prefix="repoforinstallinfo")
        populator.get_install_info(repo)

    def test_get_ordered_installable_revisions(self):
        # Used in ephemeris...
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoforindex")
        assert repository.owner
        assert repository.name
        revisions = populator.get_ordered_installable_revisions(repository.owner, repository.name)
        assert len(revisions.root) == 1

    def test_reset_on_repository(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoforreseta")
        assert repository.owner
        assert repository.name
        revisions = populator.get_ordered_installable_revisions(repository.owner, repository.name)
        assert len(revisions.root) == 1
        metadata_response = populator.reset_metadata(repository)
        assert metadata_response.start_time
        assert metadata_response.stop_time
        assert metadata_response.status == "ok"
        assert len(metadata_response.repository_status) == 1
        revisions = populator.get_ordered_installable_revisions(repository.owner, repository.name)
        assert len(revisions.root) == 1

    def test_repository_search(self):
        populator = self.populator
        repository = populator.setup_column_maker_repo(prefix="repoforreposearch")
        populator.reindex()
        results = populator.repo_search_query("repoforreposearch")
        assert len(results.hits) == 1
        first_hit = results.hits[0]
        assert first_hit.repository.name == repository.name
        assert first_hit.repository.times_downloaded == 0

    def test_repo_tars(self):
        for index, repo_path in enumerate(repo_tars("column_maker")):
            with CompressedFile(repo_path) as cf:
                path = cf.extract(tempfile.mkdtemp())
            tool_xml_path = os.path.join(path, "column_maker.xml")
            tool_source = get_tool_source(config_file=tool_xml_path)
            tool_version = tool_source.parse_version()
            if index == 0:
                assert tool_version == "1.1.0"
            elif index == 1:
                assert tool_version == "1.2.0"
            elif index == 2:
                assert tool_version == "1.3.0"
            else:
                raise AssertionError("Wrong number of repo tars returned...")

    @skip_if_api_v1
    def test_readmes(self):
        populator = self.populator
        repository = populator.setup_test_data_repo("column_maker_with_readme")
        only_revision = self._get_only_revision(repository)
        populator.assert_has_n_installable_revisions(repository, 1)
        response = self.api_interactor.get(
            f"repositories/{repository.id}/revisions/{only_revision.changeset_revision}/readmes"
        )
        api_asserts.assert_status_code_is_ok(response)
        readme_dicts = response.json()
        assert "readme.txt" in readme_dicts

    def test_reset_on_simple_repository(self):
        populator = self.populator
        repository = populator.setup_test_data_repo("column_maker")
        populator.assert_has_n_installable_revisions(repository, 3)
        response = self.api_interactor.post(
            "repositories/reset_metadata_on_repository", data={"repository_id": repository.id}
        )
        api_asserts.assert_status_code_is_ok(response)
        populator.assert_has_n_installable_revisions(repository, 3)

    def test_reset_with_uninstallable_revisions(self):
        populator = self.populator
        # setup a repository with 4 revisions but only 3 installable ones due to no version change in a tool
        repository = populator.setup_test_data_repo("column_maker_with_download_gaps")
        populator.assert_has_n_installable_revisions(repository, 3)
        response = self.api_interactor.post(
            "repositories/reset_metadata_on_repository", data={"repository_id": repository.id}
        )
        api_asserts.assert_status_code_is_ok(response)
        populator.assert_has_n_installable_revisions(repository, 3)

    @skip_if_api_v2
    def test_reset_all_v1(self):
        populator = self.populator
        repository = populator.setup_test_data_repo("column_maker_with_download_gaps")
        populator.assert_has_n_installable_revisions(repository, 3)
        # resetting one at a time or resetting everything via the web controllers works...
        # resetting all at once via the API does not work - it breaks the repository
        response = self.api_interactor.post(
            "repositories/reset_metadata_on_repositories",
            data={"payload": "can not be empty because bug in controller"},
        )
        api_asserts.assert_status_code_is_ok(response)
        populator.assert_has_n_installable_revisions(repository, 3)

    @skip_if_api_v1
    def test_reset_all_v2(self):
        populator = self.populator
        repository = populator.setup_test_data_repo("column_maker_with_download_gaps")
        populator.assert_has_n_installable_revisions(repository, 3)
        response = self.api_interactor.post("repositories/reset_metadata_on_repositories", json={})
        api_asserts.assert_status_code_is_ok(response)
        populator.assert_has_n_installable_revisions(repository, 3)

    def _get_only_revision(self, repository: HasRepositoryId) -> RepositoryRevisionMetadata:
        populator = self.populator
        repository_metadata = populator.get_metadata(repository)
        metadata_for_revisions = repository_metadata.root
        assert len(metadata_for_revisions) == 1
        only_key = list(metadata_for_revisions.keys())[0]
        assert only_key.startswith("0:")
        only_revision = list(metadata_for_revisions.values())[0]
        assert only_revision
        return only_revision



import bz2
import gzip
import json
import logging
import os
import shutil
import tempfile
from collections import namedtuple
from typing import (
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from sqlalchemy import select
from sqlalchemy.sql.expression import null

import tool_shed.repository_types.util as rt_util
from galaxy.tool_shed.tools.data_table_manager import BaseShedToolDataTableManager
from galaxy.util import checkers
from galaxy.util.path import safe_relpath
from tool_shed.util import (
    basic_util,
    hg_util,
    shed_util_common as suc,
)
from tool_shed.webapp.model import Repository

if TYPE_CHECKING:
    from tool_shed.structured_app import ToolShedApp

log = logging.getLogger(__name__)

UNDESIRABLE_DIRS = [".hg", ".svn", ".git", ".cvs", ".idea"]
UNDESIRABLE_FILES = [".hg_archival.txt", "hgrc", ".DS_Store", "tool_test_output.html", "tool_test_output.json"]


def check_archive(repository: "Repository", archive):
    valid = []
    invalid = []
    errors = []
    undesirable_files = []
    undesirable_dirs = []
    for member in archive.getmembers():
        # Allow regular files and directories only
        if not (member.isdir() or member.isfile() or member.islnk()):
            errors.append(
                "Uploaded archives can only include regular directories and files (no symbolic links, devices, etc)."
            )
            invalid.append(member)
            continue
        if not safe_relpath(member.name):
            errors.append("Uploaded archives cannot contain files that would extract outside of the archive.")
            invalid.append(member)
            continue
        if os.path.basename(member.name) in UNDESIRABLE_FILES:
            undesirable_files.append(member)
            continue
        head = tail = member.name
        found_undesirable_dir = False
        while tail:
            head, tail = os.path.split(head)
            if tail in UNDESIRABLE_DIRS:
                undesirable_dirs.append(member)
                found_undesirable_dir = True
                break
        if found_undesirable_dir:
            continue
        if (
            repository.type == rt_util.REPOSITORY_SUITE_DEFINITION
            and member.name != rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME
        ):
            errors.append(
                "Repositories of type <b>Repository suite definition</b> can contain only a single file named <b>repository_dependencies.xml</b>."
            )
            invalid.append(member)
            continue
        if (
            repository.type == rt_util.TOOL_DEPENDENCY_DEFINITION
            and member.name != rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME
        ):
            errors.append(
                "Repositories of type <b>Tool dependency definition</b> can contain only a single file named <b>tool_dependencies.xml</b>."
            )
            invalid.append(member)
            continue
        valid.append(member)
    ArchiveCheckResults = namedtuple(
        "ArchiveCheckResults", ["valid", "invalid", "undesirable_files", "undesirable_dirs", "errors"]
    )
    return ArchiveCheckResults(valid, invalid, undesirable_files, undesirable_dirs, errors)


def check_file_contents_for_email_alerts(app: "ToolShedApp"):
    """
    See if any admin users have chosen to receive email alerts when a repository is updated.
    If so, the file contents of the update must be checked for inappropriate content.
    """
    sa_session = app.model.session
    admin_users = app.config.get("admin_users", "").split(",")
    for repository in get_repositories_with_alerts(sa_session, Repository):
        email_alerts = json.loads(repository.email_alerts)
        for user_email in email_alerts:
            if user_email in admin_users:
                return True
    return False


def check_file_content_for_html_and_images(file_path):
    message = ""
    if checkers.check_html(file_path):
        message = f'The file "{str(file_path)}" contains HTML content.\n'
    elif checkers.check_image(file_path):
        message = f'The file "{str(file_path)}" contains image content.\n'
    return message


def get_change_lines_in_file_for_tag(tag, change_dict):
    """
    The received change_dict is the jsonified version of the changes to a file in a
    changeset being pushed to the Tool Shed from the command line. This method cleans
    and returns appropriate lines for inspection.
    """
    cleaned_lines = []
    data_list = change_dict.get("data", [])
    for data_dict in data_list:
        block = data_dict.get("block", "")
        lines = block.split("\\n")
        for line in lines:
            index = line.find(tag)
            if index > -1:
                line = line[index:]
                cleaned_lines.append(line)
    return cleaned_lines


def handle_bz2(repository: "Repository", uploaded_file_name):
    with (
        tempfile.NamedTemporaryFile(
            mode="wb",
            prefix=f"repo_{repository.id}_upload_bunzip2_",
            dir=os.path.dirname(uploaded_file_name),
            delete=False,
        ) as uncompressed,
        bz2.BZ2File(uploaded_file_name, "rb") as bzipped_file,
    ):
        while 1:
            try:
                chunk = bzipped_file.read(basic_util.CHUNK_SIZE)
            except OSError:
                os.remove(uncompressed.name)
                log.exception(f'Problem uncompressing bz2 data "{uploaded_file_name}"')
                return
            if not chunk:
                break
            uncompressed.write(chunk)
    shutil.move(uncompressed.name, uploaded_file_name)


ChangeResponseT = Tuple[Union[bool, str], str, List[str], str, int, int]


def handle_directory_changes(
    app: "ToolShedApp",
    host: str,
    username: str,
    repository: "Repository",
    full_path: str,
    filenames_in_archive,
    remove_repo_files_not_in_tar,
    new_repo_alert,
    commit_message: str,
    undesirable_dirs_removed: int,
    undesirable_files_removed: int,
    repo_path: Optional[str] = None,
    dry_run: bool = False,
) -> ChangeResponseT:
    repo_path = repo_path or repository.repo_path(app)
    content_alert_str = ""
    files_to_remove = []
    filenames_in_archive = [os.path.normpath(os.path.join(full_path, name)) for name in filenames_in_archive]
    if remove_repo_files_not_in_tar and not repository.is_new():
        # We have a repository that is not new (it contains files), so discover those files that are in the
        # repository, but not in the uploaded archive.
        for root, dirs, files in os.walk(full_path):
            if root.find(".hg") < 0 and root.find("hgrc") < 0:
                for undesirable_dir in UNDESIRABLE_DIRS:
                    if undesirable_dir in dirs:
                        dirs.remove(undesirable_dir)
                        undesirable_dirs_removed += 1
                for undesirable_file in UNDESIRABLE_FILES:
                    if undesirable_file in files:
                        files.remove(undesirable_file)
                        undesirable_files_removed += 1
                for name in files:
                    full_name = os.path.join(root, name)
                    if full_name not in filenames_in_archive:
                        files_to_remove.append(full_name)
        for repo_file in files_to_remove:
            # Remove files in the repository (relative to the upload point) that are not in
            # the uploaded archive.
            try:
                hg_util.remove_path(repo_path, repo_file)
            except Exception as e:
                error_message = f"Error removing file {repo_file} in mercurial repo:\n{e}"
                log.debug(error_message)
                return "error", error_message, files_to_remove, content_alert_str, 0, 0
    # See if any admin users have chosen to receive email alerts when a repository is updated.
    # If so, check every uploaded file to ensure content is appropriate.
    check_contents = check_file_contents_for_email_alerts(app)
    for filename_in_archive in filenames_in_archive:
        # Check file content to ensure it is appropriate.
        if check_contents and os.path.isfile(filename_in_archive):
            content_alert_str += check_file_content_for_html_and_images(filename_in_archive)
        hg_util.add_changeset(repo_path, filename_in_archive)
        if filename_in_archive.endswith("tool_data_table_conf.xml.sample"):
            # Handle the special case where a tool_data_table_conf.xml.sample file is being uploaded
            # by parsing the file and adding new entries to the in-memory app.tool_data_tables
            # dictionary.
            stdtm = BaseShedToolDataTableManager(app)
            error, message = stdtm.handle_sample_tool_data_table_conf_file(filename_in_archive, persist=False)
            if error:
                return (
                    False,
                    message,
                    files_to_remove,
                    content_alert_str,
                    undesirable_dirs_removed,
                    undesirable_files_removed,
                )
    hg_util.commit_changeset(repo_path, full_path_to_changeset=full_path, username=username, message=commit_message)
    admin_only = len(repository.downloadable_revisions) != 1
    if not dry_run:
        suc.handle_email_alerts(
            app,
            host,
            repository,
            content_alert_str=content_alert_str,
            new_repo_alert=new_repo_alert,
            admin_only=admin_only,
        )
    return True, "", files_to_remove, content_alert_str, undesirable_dirs_removed, undesirable_files_removed


def handle_gzip(repository, uploaded_file_name):
    with (
        tempfile.NamedTemporaryFile(
            mode="wb",
            prefix=f"repo_{repository.id}_upload_gunzip_",
            dir=os.path.dirname(uploaded_file_name),
            delete=False,
        ) as uncompressed,
        gzip.GzipFile(uploaded_file_name, "rb") as gzipped_file,
    ):
        while 1:
            try:
                chunk = gzipped_file.read(basic_util.CHUNK_SIZE)
            except OSError:
                os.remove(uncompressed.name)
                log.exception(f'Problem uncompressing gz data "{uploaded_file_name}"')
                return
            if not chunk:
                break
            uncompressed.write(chunk)
    shutil.move(uncompressed.name, uploaded_file_name)


def uncompress(repository, uploaded_file_name, uploaded_file_filename, isgzip=False, isbz2=False):
    if isgzip:
        handle_gzip(repository, uploaded_file_name)
        return uploaded_file_filename.rstrip(".gz")
    if isbz2:
        handle_bz2(repository, uploaded_file_name)
        return uploaded_file_filename.rstrip(".bz2")


def get_repositories_with_alerts(session, repository_model):
    stmt = select(repository_model).where(repository_model.email_alerts != null())
    return session.scalars(stmt)



import logging
from typing import List

from sqlalchemy import (
    false,
    select,
)

import galaxy.model
import galaxy.model.tool_shed_install
import tool_shed.webapp.model as model
from galaxy.model.db.user import (
    get_user_by_email,
    get_user_by_username,
)

log = logging.getLogger("test.tool_shed.test_db_util")


def sa_session():
    from .driver import tool_shed_context as sa_session

    return sa_session


def install_session():
    from galaxy_test.driver.driver_util import install_context as install_session

    return install_session


def flush(obj):
    sa_session().add(obj)
    sa_session().flush()


def get_all_repositories():
    return sa_session().scalars(select(model.Repository)).all()


def get_all_installed_repositories(session=None) -> List[galaxy.model.tool_shed_install.ToolShedRepository]:
    if session is None:
        session = install_session()
    ToolShedRepository = galaxy.model.tool_shed_install.ToolShedRepository
    stmt = (
        select(ToolShedRepository)
        .where(ToolShedRepository.deleted == false())
        .where(ToolShedRepository.uninstalled == false())
        .where(ToolShedRepository.status == ToolShedRepository.installation_status.INSTALLED)
    )
    return session.scalars(stmt).all()


def get_installed_repository_by_id(repository_id):
    return install_session().get(galaxy.model.tool_shed_install.ToolShedRepository, repository_id)


def get_installed_repository_by_name_owner(repository_name, owner, return_multiple=False, session=None):
    if session is None:
        session = install_session()
    ToolShedRepository = galaxy.model.tool_shed_install.ToolShedRepository
    stmt = (
        select(ToolShedRepository)
        .where(ToolShedRepository.name == repository_name)
        .where(ToolShedRepository.owner == owner)
    )
    if return_multiple:
        return session.scalars(stmt).all()
    return session.scalars(stmt.limit(1)).first()


def get_role(user, role_name):
    for role in user.all_roles():
        if role.name == role_name:
            return role
    return None


def get_repository_role_association(repository_id, role_id):
    stmt = (
        select(model.RepositoryRoleAssociation)
        .where(model.RepositoryRoleAssociation.role_id == role_id)
        .where(model.RepositoryRoleAssociation.repository_id == repository_id)
        .limit(1)
    )
    return sa_session().scalars(stmt).first()


def get_repository_by_id(repository_id):
    return sa_session().get(model.Repository, repository_id)


def get_user(email):
    return get_user_by_email(sa_session(), email, model.User)


def refresh(obj):
    sa_session().refresh(obj)


def ga_refresh(obj):
    install_session().refresh(obj)


def get_repository_by_name_and_owner(name, owner_username, return_multiple=False):
    owner = get_user_by_username(sa_session(), owner_username, model.User)
    stmt = (
        select(model.Repository)
        .where(model.Repository.name == name)
        .where(model.Repository.user_id == owner.id)
        .limit(1)
    )
    return sa_session().scalars(stmt).first()


def get_repository_metadata_by_repository_id_changeset_revision(repository_id, changeset_revision):
    stmt = (
        select(model.RepositoryMetadata)
        .where(model.RepositoryMetadata.repository_id == repository_id)
        .where(model.RepositoryMetadata.changeset_revision == changeset_revision)
        .limit(1)
    )
    return sa_session().scalars(stmt).first()



import os

from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

repository_name = "freebayes_0010"
repository_description = "Galaxy's freebayes tool"
repository_long_description = "Long description of Galaxy's freebayes tool"

"""
1. Create repository freebayes_0020 and upload only the tool XML.
2. Upload the tool_data_table_conf.xml.sample file.
3. Upload sam_fa_indices.loc.sample.
4. Upload a tool_dependencies.xml file that should not parse correctly.
5. Upload a tool_dependencies.xml file that specifies a version that does not match the tool's requirements.
6. Upload a valid tool_dependencies.xml file.
7. Check for the appropriate strings on the manage repository page.
"""


class TestFreebayesRepository(ShedTwillTestCase):
    """Testing freebayes with tool data table entries, .loc files, and tool dependencies."""

    def test_0000_create_or_login_admin_user(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_category(self):
        """Create a category for this test suite"""
        self.create_category(
            name="Test 0010 Repository With Tool Dependencies",
            description="Tests for a repository with tool dependencies.",
        )

    def test_0010_create_freebayes_repository_and_upload_tool_xml(self):
        """Create freebayes repository and upload only freebayes.xml.

        We are at step 1 - Create repository freebayes_0020 and upload only the tool XML.
        Uploading only the tool XML file should result in an invalid tool and an error message on
        upload, as well as on the manage repository page.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        category = self.populator.get_category_with_name("Test 0010 Repository With Tool Dependencies")
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        strings_displayed = ["Metadata may have been defined", "This file requires an entry", "tool_data_table_conf"]
        self.add_file_to_repository(repository, "freebayes/freebayes.xml", strings_displayed=strings_displayed)
        if self.is_v2:
            # opps... not good right?
            self.populator.reset_metadata(repository)
        self.display_manage_repository_page(
            repository, strings_displayed=[self.invalid_tools_labels], strings_not_displayed=["Valid tools"]
        )
        tip = self.get_repository_tip(repository)
        strings_displayed = ["requires an entry", "tool_data_table_conf.xml"]
        self.check_repository_invalid_tools_for_changeset_revision(repository, tip, strings_displayed=strings_displayed)

    def test_0015_upload_missing_tool_data_table_conf_file(self):
        """Upload the missing tool_data_table_conf.xml.sample file to the repository.

        We are at step 2 - Upload the tool_data_table_conf.xml.sample file.
        Uploading the tool_data_table_conf.xml.sample alone should not make the tool valid, but the error message should change.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        strings_displayed = ["Upload a file named <b>sam_fa_indices.loc.sample"]
        self.add_file_to_repository(
            repository, "freebayes/tool_data_table_conf.xml.sample", strings_displayed=strings_displayed
        )
        self.display_manage_repository_page(
            repository, strings_displayed=[self.invalid_tools_labels], strings_not_displayed=["Valid tools"]
        )
        tip = self.get_repository_tip(repository)
        strings_displayed = ["refers to a file", "sam_fa_indices.loc"]
        self.check_repository_invalid_tools_for_changeset_revision(repository, tip, strings_displayed=strings_displayed)

    def test_0020_upload_missing_sample_loc_file(self):
        """Upload the missing sam_fa_indices.loc.sample file to the repository.

        We are at step 3 - Upload the tool_data_table_conf.xml.sample file.
        Uploading the tool_data_table_conf.xml.sample alone should not make the tool valid, but the error message should change.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "freebayes/sam_fa_indices.loc.sample")

    def test_0025_upload_malformed_tool_dependency_xml(self):
        """Upload tool_dependencies.xml with bad characters in the readme tag.

        We are at step 4 - Upload a tool_dependencies.xml file that should not parse correctly.
        Upload a tool_dependencies.xml file that contains <> in the text of the readme tag. This should show an error message about malformed xml.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        target = os.path.join("freebayes", "malformed_tool_dependencies", "tool_dependencies.xml")
        self.add_file_to_repository(
            repository, target, strings_displayed=["Exception attempting to parse", "invalid element name"]
        )

    def test_0030_upload_invalid_tool_dependency_xml(self):
        """Upload tool_dependencies.xml defining version 0.9.5 of the freebayes package.

        We are at step 5 - Upload a tool_dependencies.xml file that specifies a version that does not match the tool's requirements.
        This should result in a message about the tool dependency configuration not matching the tool's requirements.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        target = os.path.join("freebayes", "invalid_tool_dependencies", "tool_dependencies.xml")
        strings_displayed = [
            "The settings for <b>name</b>, <b>version</b> and <b>type</b> from a contained tool configuration"
        ]
        self.add_file_to_repository(repository, target, strings_displayed=strings_displayed)

    def test_0035_upload_valid_tool_dependency_xml(self):
        """Upload tool_dependencies.xml defining version 0.9.4_9696d0ce8a962f7bb61c4791be5ce44312b81cf8 of the freebayes package.

        We are at step 6 - Upload a valid tool_dependencies.xml file.
        At this stage, there should be no errors on the upload page, as every missing or invalid file has been corrected.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        target = os.path.join("freebayes", "tool_dependencies.xml")
        self.add_file_to_repository(repository, target)

    @skip_if_api_v2
    def test_0040_verify_tool_dependencies(self):
        """Verify that the uploaded tool_dependencies.xml specifies the correct package versions.

        We are at step 7 - Check for the appropriate strings on the manage repository page.
        Verify that the manage repository page now displays the valid tool dependencies, and that there are no invalid tools shown on the manage page.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        strings_displayed = ["freebayes", "0.9.4_9696d0ce8a9", "samtools", "0.1.18", "Valid tools", "package"]
        strings_not_displayed = [self.invalid_tools_labels]
        self.display_manage_repository_page(
            repository, strings_displayed=strings_displayed, strings_not_displayed=strings_not_displayed
        )



"""
Manager and Serializer for TS groups.
"""

import logging

from sqlalchemy import (
    false,
    select,
    true,
)
from sqlalchemy.exc import (
    MultipleResultsFound,
    NoResultFound,
)

from galaxy.exceptions import (
    Conflict,
    InconsistentDatabase,
    InternalServerError,
    ItemAccessibilityException,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from tool_shed.webapp.model import Group

log = logging.getLogger(__name__)


class GroupManager:
    """
    Interface/service object for interacting with TS groups.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, trans, decoded_group_id=None, name=None):
        """
        Get the group from the DB based on its ID or name.

        :param  decoded_group_id:       decoded group id
        :type   decoded_group_id:       int

        :returns:   the requested group
        :rtype:     tool_shed.model.Group
        """
        if decoded_group_id is None and name is None:
            raise RequestParameterInvalidException("You must supply either ID or a name of the group.")

        try:
            if decoded_group_id:
                group = trans.sa_session.get(Group, decoded_group_id)
            else:
                group = get_group_by_name(trans.sa_session, name, Group)
        except MultipleResultsFound:
            raise InconsistentDatabase("Multiple groups found with the same identifier.")
        except NoResultFound:
            raise ObjectNotFound("No group found with the identifier provided.")
        except Exception:
            raise InternalServerError("Error loading from the database.")
        return group

    def create(self, trans, name, description=""):
        """
        Create a new group.
        """
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can create groups.")
        else:
            if self.get(trans, name=name):
                raise Conflict(f"Group with the given name already exists. Name: {str(name)}")
            # TODO add description field to the model
            group = Group(name=name)
            trans.sa_session.add(group)
            trans.sa_session.commit()
            return group

    def update(self, trans, group, name=None, description=None):
        """
        Update the given group
        """
        changed = False
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can update groups.")
        if group.deleted:
            raise RequestParameterInvalidException("You cannot modify a deleted group. Undelete it first.")
        if name is not None:
            group.name = name
            changed = True
        if description is not None:
            group.description = description
            changed = True
        if changed:
            trans.sa_session.add(group)
            trans.sa_session.commit()
        return group

    def delete(self, trans, group, undelete=False):
        """
        Mark given group deleted/undeleted based on the flag.
        """
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can delete and undelete groups.")
        if undelete:
            group.deleted = False
        else:
            group.deleted = True
        trans.sa_session.add(group)
        trans.sa_session.commit()
        return group

    def list(self, trans, deleted=False):
        """
        Return a list of groups from the DB.

        :returns: query that will emit all groups
        :rtype:   sqlalchemy query
        """
        stmt = select(Group)
        if trans.user_is_admin:
            if deleted is None:
                #  Flag is not specified, do not filter on it.
                pass
            elif deleted:
                stmt = stmt.where(Group.deleted == true())
            else:
                stmt = stmt.where(Group.deleted == false())
        else:
            stmt = stmt.where(Group.deleted == false())
        return trans.sa_session.scalars(stmt)


def get_group_by_name(session, name, group_model):
    stmt = select(group_model).where(group_model.name == name)
    return session.execute(stmt).scalar_one()



import logging
import os

from mercurial import (
    hg,
    ui,
)
from sqlalchemy import (
    false,
    select,
)
from whoosh.writing import AsyncWriter

import tool_shed.webapp.model.mapping as ts_mapping
from galaxy.tool_util.loader_directory import load_tool_elements_from_path
from galaxy.tools.search import get_or_create_index
from galaxy.util import (
    directory_hash_id,
    ExecutionTimer,
    pretty_print_time_interval,
    unicodify,
)
from tool_shed.util.hgweb_config import hgweb_config_manager
from tool_shed.webapp import model
from tool_shed.webapp.search.repo_search import schema as repo_schema
from tool_shed.webapp.search.tool_search import schema as tool_schema

log = logging.getLogger(__name__)


def _get_or_create_index(whoosh_index_dir):
    tool_index_dir = os.path.join(whoosh_index_dir, "tools")
    if not os.path.exists(whoosh_index_dir):
        os.makedirs(whoosh_index_dir)
    if not os.path.exists(tool_index_dir):
        os.makedirs(tool_index_dir)
    return get_or_create_index(whoosh_index_dir, repo_schema), get_or_create_index(tool_index_dir, tool_schema)


def build_index(whoosh_index_dir, file_path, hgweb_config_dir, hgweb_repo_prefix, dburi, **kwargs):
    """
    Build two search indexes simultaneously
    One is for repositories and the other for tools.

    Returns a tuple with number of repos and tools that were indexed.
    """
    model = ts_mapping.init(dburi, engine_options={}, create_tables=False)
    sa_session = model.session
    repo_index, tool_index = _get_or_create_index(whoosh_index_dir)

    repo_index_writer = AsyncWriter(repo_index)
    tool_index_writer = AsyncWriter(tool_index)
    repos_indexed = 0
    tools_indexed = 0

    execution_timer = ExecutionTimer()
    with repo_index.searcher() as searcher:
        for repo in get_repos(sa_session, file_path, hgweb_config_dir, hgweb_repo_prefix, **kwargs):
            if repo is None:
                continue
            tools_list = repo.pop("tools_list")
            repo_id = repo["id"]
            indexed_document = searcher.document(id=repo_id)
            if indexed_document:
                if indexed_document["full_last_updated"] == repo.get("full_last_updated"):
                    # We're done, since we sorted repos by update time
                    break
                else:
                    # Got an update, delete the previous document
                    repo_index_writer.delete_by_term("id", repo_id)

            repo_index_writer.add_document(**repo)

            #  Tools get their own index
            tool_index_writer.delete_by_term("repo_id", repo_id)
            for tool in tools_list:
                tool_contents = tool.copy()
                tool_contents["repo_owner_username"] = repo.get("repo_owner_username")
                tool_contents["repo_name"] = repo.get("name")
                tool_contents["repo_id"] = repo_id
                tool_index_writer.add_document(**tool_contents)
                tools_indexed += 1

            repos_indexed += 1

    tool_index_writer.commit()
    repo_index_writer.commit()

    log.info("Indexed repos: %s, tools: %s", repos_indexed, tools_indexed)
    log.info("Toolbox index finished %s", execution_timer)
    return repos_indexed, tools_indexed


def get_repos(sa_session, file_path, hgweb_config_dir, hgweb_repo_prefix, **kwargs):
    """
    Load repos from DB and included tools from .xml configs.
    """
    hgwcm = hgweb_config_manager
    hgwcm.hgweb_config_dir = hgweb_config_dir
    for repo in get_repositories_for_indexing(sa_session):
        category_names = []
        for rca in get_repo_cat_associations(sa_session, repo.id):
            category = sa_session.get(model.Category, rca.category.id)
            category_names.append(category.name.lower())
        categories = (",").join(category_names)
        repo_id = repo.id
        name = repo.name
        description = repo.description
        long_description = repo.long_description
        homepage_url = repo.homepage_url
        remote_repository_url = repo.remote_repository_url

        times_downloaded = repo.times_downloaded or 0

        repo_owner_username = ""
        if repo.user_id is not None:
            user = sa_session.get(model.User, repo.user_id)
            repo_owner_username = user.username.lower()

        last_updated = pretty_print_time_interval(repo.update_time)
        full_last_updated = repo.update_time.strftime("%Y-%m-%d %I:%M %p")

        # Load all changesets of the repo for lineage.
        try:
            entry = hgwcm.get_entry(os.path.join(hgweb_repo_prefix, repo.user.username, repo.name))
        except Exception:
            return None
        repo_path = os.path.join(hgweb_config_dir, entry)
        hg_repo = hg.repository(ui.ui(), repo_path.encode("utf-8"))
        lineage = []
        for changeset in hg_repo.changelog:
            lineage.append(f"{unicodify(changeset)}:{unicodify(hg_repo[changeset])}")
        repo_lineage = str(lineage)

        #  Parse all the tools within repo for a separate index.
        tools_list = []
        path = os.path.join(file_path, *directory_hash_id(repo.id))
        path = os.path.join(path, f"repo_{repo.id}")
        if os.path.exists(path):
            tools_list.extend(load_one_dir(path))
            for root, dirs, _files in os.walk(path):
                if ".hg" in dirs:
                    dirs.remove(".hg")
                for dirname in dirs:
                    tools_in_dir = load_one_dir(os.path.join(root, dirname))
                    tools_list.extend(tools_in_dir)

        yield (
            dict(
                id=unicodify(repo_id),
                name=unicodify(name),
                description=unicodify(description),
                long_description=unicodify(long_description),
                homepage_url=unicodify(homepage_url),
                remote_repository_url=unicodify(remote_repository_url),
                repo_owner_username=unicodify(repo_owner_username),
                times_downloaded=unicodify(times_downloaded),
                approved=unicodify("no"),
                last_updated=unicodify(last_updated),
                full_last_updated=unicodify(full_last_updated),
                tools_list=tools_list,
                repo_lineage=unicodify(repo_lineage),
                categories=unicodify(categories),
            )
        )


def debug_handler(path, exc_info):
    """
    By default the underlying tool parsing logs warnings for each exception.
    This is very chatty hence this metod changes it to debug level.
    """
    log.debug(f"Failed to load tool with path {path}.", exc_info=exc_info)


def load_one_dir(path):
    tools_in_dir = []
    tool_elems = load_tool_elements_from_path(path, load_exception_handler=debug_handler)
    if tool_elems:
        for elem in tool_elems:
            root = elem[1].getroot()
            if root.tag == "tool":
                tool = {}
                if root.find("help") is not None:
                    tool.update(dict(help=unicodify(root.find("help").text)))
                if root.find("description") is not None:
                    tool.update(dict(description=unicodify(root.find("description").text)))
                tool.update(
                    dict(
                        id=unicodify(root.attrib.get("id")),
                        name=unicodify(root.attrib.get("name")),
                        version=unicodify(root.attrib.get("version")),
                    )
                )
                tools_in_dir.append(tool)
    return tools_in_dir


def get_repositories_for_indexing(session):
    # Do not index deleted, deprecated, or "tool_dependency_definition" type repositories.
    Repository = model.Repository
    stmt = (
        select(Repository)
        .where(Repository.deleted == false())
        .where(Repository.deprecated == false())
        .where(Repository.type != "tool_dependency_definition")
        .order_by(Repository.update_time.desc())
    )
    return session.scalars(stmt)


def get_repo_cat_associations(session, repository_id):
    stmt = select(model.RepositoryCategoryAssociation).where(
        model.RepositoryCategoryAssociation.repository_id == repository_id
    )
    return session.scalars(stmt)



import os
import tempfile
from collections import namedtuple
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from galaxy import exceptions
from galaxy.exceptions import (
    InconsistentApplicationState,
    InternalServerError,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.tool_shed.metadata.metadata_generator import RepositoryMetadataToolDict
from galaxy.tool_shed.util.basic_util import remove_dir
from galaxy.tool_shed.util.hg_util import (
    clone_repository,
    get_changectx_for_changeset,
)
from galaxy.tool_util.model_factory import parse_tool_custom
from galaxy.tool_util.parser import (
    get_tool_source,
    ToolSource,
)
from galaxy.tools.stock import stock_tool_sources
from galaxy.util import relpath
from tool_shed.context import (
    ProvidesRepositoriesContext,
    SessionRequestContext,
)
from tool_shed.util.common_util import generate_clone_url_for
from tool_shed.webapp.model import RepositoryMetadata
from tool_shed.webapp.search.tool_search import ToolSearch
from tool_shed_client.schema import ShedParsedTool
from .repositories import get_repository_revision_metadata_model
from .trs import trs_tool_id_to_repository_metadata

STOCK_TOOL_SOURCES: Optional[Dict[str, Dict[str, ToolSource]]] = None


def search(trans: SessionRequestContext, q: str, page: int = 1, page_size: int = 10) -> dict:
    """
    Perform the search over TS tools index.
    Note that search works over the Whoosh index which you have
    to pre-create with scripts/tool_shed/build_ts_whoosh_index.sh manually.
    Also TS config option toolshed_search_on has to be True and
    whoosh_index_dir has to be specified.
    """
    app = trans.app
    conf = app.config
    if not conf.toolshed_search_on:
        raise exceptions.ConfigDoesNotAllowException(
            "Searching the TS through the API is turned off for this instance."
        )
    if not conf.whoosh_index_dir:
        raise exceptions.ConfigDoesNotAllowException(
            "There is no directory for the search index specified. Please contact the administrator."
        )
    search_term = q.strip()
    if len(search_term) < 1:
        raise exceptions.RequestParameterInvalidException("The search term has to be at least one character long.")

    tool_search = ToolSearch()

    Boosts = namedtuple(
        "Boosts", ["tool_name_boost", "tool_description_boost", "tool_help_boost", "tool_repo_owner_username_boost"]
    )
    boosts = Boosts(
        float(conf.get("tool_name_boost", 1.2)),
        float(conf.get("tool_description_boost", 0.6)),
        float(conf.get("tool_help_boost", 0.4)),
        float(conf.get("tool_repo_owner_username_boost", 0.3)),
    )

    results = tool_search.search(trans.app, search_term, page, page_size, boosts)
    results["hostname"] = trans.repositories_hostname
    return results


def get_repository_metadata_tool_dict(
    trans: ProvidesRepositoriesContext, trs_tool_id: str, tool_version: str
) -> Tuple[RepositoryMetadata, RepositoryMetadataToolDict]:
    if trs_tool_id.count("~") < 2:
        RequestParameterInvalidException(f"Invalid TRS tool id ({trs_tool_id})")

    name, owner, tool_id = trs_tool_id.split("~", 3)
    repository, metadata_by_version = trs_tool_id_to_repository_metadata(trans, trs_tool_id)
    if tool_version not in metadata_by_version:
        raise ObjectNotFound()
    tool_version_repository_metadata: RepositoryMetadata = metadata_by_version[tool_version]
    raw_metadata = tool_version_repository_metadata.metadata
    tool_dicts: List[RepositoryMetadataToolDict] = raw_metadata.get("tools", [])
    for tool_dict in tool_dicts:
        if tool_dict["id"] != tool_id or tool_dict["version"] != tool_version:
            continue
        return tool_version_repository_metadata, tool_dict
    raise ObjectNotFound()


def parsed_tool_model_cached_for(
    trans: ProvidesRepositoriesContext, trs_tool_id: str, tool_version: str, repository_clone_url: Optional[str] = None
) -> ShedParsedTool:
    model_cache = trans.app.model_cache
    parsed_tool = model_cache.get_cache_entry_for(ShedParsedTool, trs_tool_id, tool_version)
    if parsed_tool is not None:
        return parsed_tool
    parsed_tool = parsed_tool_model_for(trans, trs_tool_id, tool_version, repository_clone_url=repository_clone_url)
    model_cache.insert_cache_entry_for(parsed_tool, trs_tool_id, tool_version)
    return parsed_tool


def parsed_tool_model_for(
    trans: ProvidesRepositoriesContext, trs_tool_id: str, tool_version: str, repository_clone_url: Optional[str] = None
) -> ShedParsedTool:
    tool_source, repository_metadata = tool_source_for(
        trans, trs_tool_id, tool_version, repository_clone_url=repository_clone_url
    )
    parsed_tool = parse_tool_custom(tool_source, ShedParsedTool)
    if repository_metadata:
        revision_model = get_repository_revision_metadata_model(
            trans.app, repository_metadata.repository, repository_metadata, recursive=False
        )
        parsed_tool.repository_revision = revision_model
    return parsed_tool


def tool_source_for(
    trans: ProvidesRepositoriesContext, trs_tool_id: str, tool_version: str, repository_clone_url: Optional[str] = None
) -> Tuple[ToolSource, Optional[RepositoryMetadata]]:
    if "~" in trs_tool_id:
        return _shed_tool_source_for(trans, trs_tool_id, tool_version, repository_clone_url)
    else:
        tool_source = _stock_tool_source_for(trs_tool_id, tool_version)
        if tool_source is None:
            raise ObjectNotFound()
        return tool_source, None


def _shed_tool_source_for(
    trans: ProvidesRepositoriesContext, trs_tool_id: str, tool_version: str, repository_clone_url: Optional[str] = None
) -> Tuple[ToolSource, RepositoryMetadata]:
    rval = get_repository_metadata_tool_dict(trans, trs_tool_id, tool_version)
    repository_metadata, tool_version_metadata = rval
    tool_config = tool_version_metadata["tool_config"]

    repo = repository_metadata.repository.hg_repo
    ctx = get_changectx_for_changeset(repo, repository_metadata.changeset_revision)
    work_dir = tempfile.mkdtemp(prefix="tmp-toolshed-tool_source")
    if repository_clone_url is None:
        repository_clone_url = generate_clone_url_for(trans, repository_metadata.repository)
    try:
        cloned_ok, error_message = clone_repository(repository_clone_url, work_dir, str(ctx.rev()))
        if error_message:
            raise InternalServerError("Failed to materialize target repository revision")
        repo_files_dir = repository_metadata.repository.hg_repository_path(trans.app.config.file_path)
        if not repo_files_dir:
            raise InconsistentApplicationState(
                f"Failed to resolve repository path from hgweb_config_manager for [{trs_tool_id}], inconsistent repository state or application configuration"
            )
        repo_rel_tool_path = relpath(tool_config, repo_files_dir)
        path_to_tool = os.path.join(work_dir, repo_rel_tool_path)
        if not os.path.exists(path_to_tool):
            raise InconsistentApplicationState(
                f"Target tool expected at [{path_to_tool}] and not found, inconsistent repository state or application configuration"
            )
        tool_source = get_tool_source(path_to_tool)
        return tool_source, repository_metadata
    finally:
        remove_dir(work_dir)


def _stock_tool_source_for(tool_id: str, tool_version: str) -> Optional[ToolSource]:
    _init_stock_tool_sources()
    assert STOCK_TOOL_SOURCES
    tool_version_sources = STOCK_TOOL_SOURCES.get(tool_id)
    if tool_version_sources is None:
        return None
    return tool_version_sources.get(tool_version)


def _init_stock_tool_sources() -> None:
    global STOCK_TOOL_SOURCES
    if STOCK_TOOL_SOURCES is None:
        STOCK_TOOL_SOURCES = {}
        for tool_source in stock_tool_sources():
            tool_id = tool_source.parse_id()
            tool_version = tool_source.parse_version()
            if tool_id not in STOCK_TOOL_SOURCES:
                STOCK_TOOL_SOURCES[tool_id] = {}
            STOCK_TOOL_SOURCES[tool_id][tool_version] = tool_source



import copy
import logging
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

from galaxy.util import (
    asbool,
    Element,
    SubElement,
)
from tool_shed.dependencies.tool import tag_attribute_handler
from tool_shed.repository_types.util import (
    REPOSITORY_DEPENDENCY_DEFINITION_FILENAME,
    TOOL_DEPENDENCY_DEFINITION_FILENAME,
)
from tool_shed.util import (
    hg_util,
    metadata_util,
    xml_util,
)
from tool_shed.webapp.model.db import get_repository_by_name_and_owner

if TYPE_CHECKING:
    from tool_shed.context import ProvidesRepositoriesContext
    from tool_shed.structured_app import ToolShedApp


log = logging.getLogger(__name__)


class RepositoryDependencyAttributeHandler:
    trans: "ProvidesRepositoriesContext"
    app: "ToolShedApp"

    def __init__(self, trans: "ProvidesRepositoriesContext", unpopulate):
        self.trans = trans
        self.app = trans.app
        self.file_name = REPOSITORY_DEPENDENCY_DEFINITION_FILENAME
        self.unpopulate = unpopulate

    def check_tag_attributes(self, elem):
        # <repository name="molecule_datatypes" owner="test" />
        error_message = ""
        name = elem.get("name")
        if not name:
            error_message += "The tag is missing the required name attribute.  "
        owner = elem.get("owner")
        if not owner:
            error_message += "The tag is missing the required owner attribute.  "
        log.debug(error_message)
        return error_message

    def handle_complex_dependency_elem(self, parent_elem, elem_index, elem):
        """
        Populate or unpopulate the toolshed and changeset_revision attributes of a
        <repository> tag that defines a complex repository dependency.
        """
        # <repository name="package_eigen_2_0" owner="test" prior_installation_required="True" />
        altered, new_elem, error_message = self.handle_elem(elem)
        if error_message:
            error_message += f"  The {TOOL_DEPENDENCY_DEFINITION_FILENAME} file contains an invalid <repository> tag."
        return altered, new_elem, error_message

    def handle_elem(self, elem):
        """Populate or unpopulate the changeset_revision and toolshed attributes of repository tags."""
        # <repository name="molecule_datatypes" owner="test" changeset_revision="1a070566e9c6" />
        # <repository changeset_revision="xxx" name="package_xorg_macros_1_17_1" owner="test" toolshed="yyy">
        #    <package name="xorg_macros" version="1.17.1" />
        # </repository>
        error_message = ""
        name = elem.get("name")
        owner = elem.get("owner")
        # The name and owner attributes are always required, so if either are missing, return the error message.
        if not name or not owner:
            error_message = self.check_tag_attributes(elem)
            return False, elem, error_message
        altered = False
        toolshed = elem.get("toolshed")
        changeset_revision = elem.get("changeset_revision")
        # Over a short period of time a bug existed which caused the prior_installation_required attribute
        # to be set to False and included in the <repository> tag when a repository was exported along with
        # its dependencies.  The following will eliminate this problematic attribute upon import.
        prior_installation_required = elem.get("prior_installation_required")
        if prior_installation_required is not None and not asbool(prior_installation_required):
            del elem.attrib["prior_installation_required"]
        sub_elems = list(elem)
        if len(sub_elems) > 0:
            # At this point, a <repository> tag will point only to a package.
            # <package name="xorg_macros" version="1.17.1" />
            # Coerce the list to dict.
            sub_elements = {}
            packages = []
            for sub_elem in sub_elems:
                sub_elem_type = sub_elem.tag
                sub_elem_name = sub_elem.get("name")
                sub_elem_version = sub_elem.get("version")
                if sub_elem_type and sub_elem_name and sub_elem_version:
                    packages.append((sub_elem_name, sub_elem_version))
            sub_elements["packages"] = packages
        else:
            # Set to None.
            sub_elements = None
        if self.unpopulate:
            # We're exporting the repository, so eliminate all toolshed and changeset_revision attributes
            # from the <repository> tag.
            if toolshed or changeset_revision:
                attributes = {}
                attributes["name"] = name
                attributes["owner"] = owner
                prior_installation_required = elem.get("prior_installation_required")
                if asbool(prior_installation_required):
                    attributes["prior_installation_required"] = "True"
                new_elem = _create_element("repository", attributes=attributes, sub_elements=sub_elements)
                altered = True
            return altered, new_elem, error_message
        # From here on we're populating the toolshed and changeset_revision attributes if necessary.
        if not toolshed:
            # Default the setting to the current tool shed.
            toolshed = str(self.trans.url_builder("/", qualified=True)).rstrip("/")
            elem.attrib["toolshed"] = toolshed
            altered = True
        if not changeset_revision:
            # Populate the changeset_revision attribute with the latest installable metadata revision for
            # the defined repository.  We use the latest installable revision instead of the latest metadata
            # revision to ensure that the contents of the revision are valid.
            repository = get_repository_by_name_and_owner(self.app.model.context, name, owner)
            if repository:
                lastest_installable_changeset_revision = metadata_util.get_latest_downloadable_changeset_revision(
                    self.app, repository
                )
                if lastest_installable_changeset_revision != hg_util.INITIAL_CHANGELOG_HASH:
                    elem.attrib["changeset_revision"] = lastest_installable_changeset_revision
                    altered = True
                else:
                    error_message = (
                        f"Invalid latest installable changeset_revision {lastest_installable_changeset_revision} "
                    )
                    error_message += f"retrieved for repository {name} owned by {owner}.  "
            else:
                error_message = f"Unable to locate repository with name {name} and owner {owner}.  "
        return altered, elem, error_message

    def handle_sub_elem(self, parent_elem, elem_index, elem):
        """
        Populate or unpopulate the toolshed and changeset_revision attributes for each of
        the following tag sets.
        <action type="set_environment_for_install">
        <action type="setup_r_environment">
        <action type="setup_ruby_environment">
        """
        sub_elem_altered = False
        error_message = ""
        for sub_index, sub_elem in enumerate(elem):
            # Make sure to skip comments and tags that are not <repository>.
            if sub_elem.tag == "repository":
                altered, new_sub_elem, message = self.handle_elem(sub_elem)
                if message:
                    error_message += f"The {TOOL_DEPENDENCY_DEFINITION_FILENAME} file contains an invalid <repository> tag.  {message}"
                if altered:
                    if not sub_elem_altered:
                        sub_elem_altered = True
                    elem[sub_index] = new_sub_elem
        if sub_elem_altered:
            parent_elem[elem_index] = elem
        return sub_elem_altered, parent_elem, error_message

    def handle_tag_attributes(self, config):
        """
        Populate or unpopulate the toolshed and changeset_revision attributes of a
        <repository> tag.  Populating will occur when a dependency definition file
        is being uploaded to the repository, while unpopulating will occur when the
        repository is being exported.
        """
        # Make sure we're looking at a valid repository_dependencies.xml file.
        tree, error_message = xml_util.parse_xml(config)
        if tree is None:
            return False, None, error_message
        root = tree.getroot()
        root_altered = False
        new_root = copy.deepcopy(root)
        for index, elem in enumerate(root):
            if elem.tag == "repository":
                # <repository name="molecule_datatypes" owner="test" changeset_revision="1a070566e9c6" />
                altered, new_elem, error_message = self.handle_elem(elem)
                if error_message:
                    error_message = f"The {self.file_name} file contains an invalid <repository> tag.  {error_message}"
                    return False, None, error_message
                if altered:
                    if not root_altered:
                        root_altered = True
                    new_root[index] = new_elem
        return root_altered, new_root, error_message


class ToolDependencyAttributeHandler:
    trans: "ProvidesRepositoriesContext"
    app: "ToolShedApp"

    def __init__(self, trans: "ProvidesRepositoriesContext", unpopulate):
        self.trans = trans
        self.app = trans.app
        self.file_name = TOOL_DEPENDENCY_DEFINITION_FILENAME
        self.unpopulate = unpopulate

    def handle_tag_attributes(self, tool_dependencies_config):
        """
        Populate or unpopulate the tooshed and changeset_revision attributes of each <repository>
        tag defined within a tool_dependencies.xml file.
        """
        rdah = RepositoryDependencyAttributeHandler(self.trans, self.unpopulate)
        tah = tag_attribute_handler.TagAttributeHandler(self.app, rdah, self.unpopulate)
        altered = False
        error_message = ""
        # Make sure we're looking at a valid tool_dependencies.xml file.
        tree, error_message = xml_util.parse_xml(tool_dependencies_config)
        if tree is None:
            return False, None, error_message
        root = tree.getroot()
        altered, new_root, error_message = tah.process_config(root, skip_actions_tags=False)
        return altered, new_root, error_message


def _create_element(
    tag: str,
    attributes: Optional[Dict[str, str]] = None,
    sub_elements: Optional[Dict[str, List[Tuple[str, str]]]] = None,
) -> Optional[Element]:
    """
    Create a new element whose tag is the value of the received tag, and whose attributes are all
    key / value pairs in the received attributes and sub_elements.
    """
    if tag:
        elem = Element(tag)
        if attributes:
            # The received attributes is an odict to preserve ordering.
            for k, attribute_value in attributes.items():
                elem.set(k, attribute_value)
        if sub_elements:
            # The received attributes is an odict.  These handle information that tends to be
            # long text including paragraphs (e.g., description and long_description.
            for k, v in sub_elements.items():
                # Don't include fields that are blank.
                if v:
                    if k == "packages":
                        # The received sub_elements is an odict whose key is 'packages' and whose
                        # value is a list of (name, version) tuples.
                        for v_tuple in v:
                            sub_elem = SubElement(elem, "package")
                            sub_elem_name, sub_elem_version = v_tuple
                            sub_elem.set("name", sub_elem_name)
                            sub_elem.set("version", sub_elem_version)
                    elif isinstance(v, list):
                        sub_elem = SubElement(elem, k)
                        # If v is a list, then it must be a list of tuples where the first
                        # item is the tag and the second item is the text value.
                        for v_tuple in v:
                            if len(v_tuple) == 2:
                                v_tag = v_tuple[0]
                                v_text = v_tuple[1]
                                # Don't include fields that are blank.
                                if v_text:
                                    v_elem = SubElement(sub_elem, v_tag)
                                    v_elem.text = v_text
                    else:
                        sub_elem = SubElement(elem, k)
                        sub_elem.text = v
        return elem
    return None



#!/usr/bin/env python
"""Test driver for tool shed functional tests."""

import os
import string
import tempfile

# This is for the tool shed application.
from galaxy.webapps.galaxy.buildapp import app_factory as galaxy_app_factory
from galaxy_test.driver import driver_util
from tool_shed.test.base.api_util import get_admin_api_key
from tool_shed.webapp import buildapp as toolshedbuildapp
from tool_shed.webapp.app import UniverseApplication as ToolshedUniverseApplication
from tool_shed.webapp.fast_app import initialize_fast_app as init_tool_shed_fast_app

log = driver_util.build_logger()

tool_sheds_conf_xml_template = """<?xml version="1.0"?>
<tool_sheds>
    <tool_shed name="Galaxy main tool shed" url="http://toolshed.g2.bx.psu.edu/"/>
    <tool_shed name="Galaxy test tool shed" url="http://testtoolshed.g2.bx.psu.edu/"/>
    <tool_shed name="Embedded tool shed for functional tests" url="http://${shed_url}:${shed_port}/"/>
</tool_sheds>
"""

shed_tool_conf_xml_template = """<?xml version="1.0"?>
<toolbox tool_path="${shed_tool_path}">
</toolbox>
"""

tool_data_table_conf_xml_template = """<?xml version="1.0"?>
<tables>
</tables>
"""

shed_data_manager_conf_xml_template = """<?xml version="1.0"?>
<data_managers>
</data_managers>
"""
# Global variable to pass database contexts around - only needed for older
# Tool Shed twill tests that didn't utilize the API for such interactions.
tool_shed_context = None


def build_shed_app(simple_kwargs):
    """Build a Galaxy app object from a simple keyword arguments.

    Construct paste style complex dictionary. Also setup "global" reference
    to sqlalchemy database context for tool shed database.
    """
    log.info("Tool shed database connection: %s", simple_kwargs["database_connection"])
    # TODO: Simplify global_conf to match Galaxy above...
    simple_kwargs["__file__"] = "tool_shed_wsgi.yml.sample"
    simple_kwargs["global_conf"] = driver_util.get_webapp_global_conf()

    app = ToolshedUniverseApplication(**simple_kwargs)
    log.info("Embedded Toolshed application started")

    global tool_shed_context
    tool_shed_context = app.model.context

    return app


class ToolShedTestDriver(driver_util.TestDriver):
    """Instantiate a Galaxy-style TestDriver for testing the tool shed."""

    def setup(self) -> None:
        """Entry point for test driver script."""
        self.external_shed = bool(os.environ.get("TOOL_SHED_TEST_EXTERNAL", None))
        if not self.external_shed:
            self._setup_local()
        else:
            # Going to also need to set TOOL_SHED_TEST_HOST.
            assert os.environ["TOOL_SHED_TEST_HOST"]

    def _setup_local(self):
        # ---- Configuration ------------------------------------------------------

        # disable sensitive request limiting for the test suite
        os.environ["TOOL_SHED_SENSITIVE_API_REQUEST_LIMIT"] = "10000/second"

        tool_shed_test_tmp_dir = driver_util.setup_tool_shed_tmp_dir()
        if not os.path.isdir(tool_shed_test_tmp_dir):
            os.mkdir(tool_shed_test_tmp_dir)
        self.temp_directories.append(tool_shed_test_tmp_dir)
        shed_db_path = driver_util.database_files_path(tool_shed_test_tmp_dir, prefix="TOOL_SHED")
        shed_tool_data_table_conf_file = os.environ.get(
            "TOOL_SHED_TEST_TOOL_DATA_TABLE_CONF", os.path.join(tool_shed_test_tmp_dir, "shed_tool_data_table_conf.xml")
        )
        galaxy_shed_data_manager_conf_file = os.environ.get(
            "GALAXY_SHED_DATA_MANAGER_CONF", os.path.join(tool_shed_test_tmp_dir, "test_shed_data_manager_conf.xml")
        )
        default_tool_data_table_config_path = os.path.join(tool_shed_test_tmp_dir, "tool_data_table_conf.xml")
        galaxy_shed_tool_conf_file = os.environ.get(
            "GALAXY_TEST_SHED_TOOL_CONF", os.path.join(tool_shed_test_tmp_dir, "test_shed_tool_conf.xml")
        )
        galaxy_migrated_tool_conf_file = os.environ.get(
            "GALAXY_TEST_MIGRATED_TOOL_CONF", os.path.join(tool_shed_test_tmp_dir, "test_migrated_tool_conf.xml")
        )
        galaxy_tool_sheds_conf_file = os.environ.get(
            "GALAXY_TEST_TOOL_SHEDS_CONF", os.path.join(tool_shed_test_tmp_dir, "test_sheds_conf.xml")
        )
        if "GALAXY_TEST_TOOL_DATA_PATH" in os.environ:
            tool_data_path = os.environ["GALAXY_TEST_TOOL_DATA_PATH"]
        else:
            tool_data_path = tempfile.mkdtemp(dir=tool_shed_test_tmp_dir)
            os.environ["GALAXY_TEST_TOOL_DATA_PATH"] = tool_data_path
        galaxy_db_path = driver_util.database_files_path(tool_shed_test_tmp_dir)
        shed_file_path = os.path.join(shed_db_path, "files")
        whoosh_index_dir = os.path.join(shed_db_path, "toolshed_whoosh_indexes")
        hgweb_config_dir = tempfile.mkdtemp(dir=tool_shed_test_tmp_dir)
        new_repos_path = tempfile.mkdtemp(dir=tool_shed_test_tmp_dir)
        galaxy_shed_tool_path = tempfile.mkdtemp(dir=tool_shed_test_tmp_dir)
        galaxy_migrated_tool_path = tempfile.mkdtemp(dir=tool_shed_test_tmp_dir)
        os.environ["TEST_HG_WEB_CONFIG_DIR"] = hgweb_config_dir
        print("Directory location for hgweb.config:", hgweb_config_dir)
        toolshed_database_conf = driver_util.database_conf(shed_db_path, prefix="TOOL_SHED")
        kwargs = dict(
            admin_users="test@bx.psu.edu",
            bootstrap_admin_api_key=get_admin_api_key(),
            allow_user_creation=True,
            allow_user_deletion=True,
            datatype_converters_config_file="datatype_converters_conf.xml.sample",
            file_path=shed_file_path,
            hgweb_config_dir=hgweb_config_dir,
            id_secret="changethisinproductiontoo",
            log_destination="stdout",
            new_file_path=new_repos_path,
            running_functional_tests=True,
            shed_tool_data_table_config=shed_tool_data_table_conf_file,
            smtp_server="smtp.dummy.string.tld",
            email_from="functional@localhost",
            use_heartbeat=False,
            whoosh_index_dir=whoosh_index_dir,
        )
        kwargs.update(toolshed_database_conf)
        # Generate the tool_data_table_conf.xml file.
        with open(default_tool_data_table_config_path, "w") as fh:
            fh.write(tool_data_table_conf_xml_template)
        # Generate the shed_tool_data_table_conf.xml file.
        with open(shed_tool_data_table_conf_file, "w") as fh:
            fh.write(tool_data_table_conf_xml_template)
        os.environ["TOOL_SHED_TEST_TOOL_DATA_TABLE_CONF"] = shed_tool_data_table_conf_file
        # ---- Run tool shed webserver ------------------------------------------------------
        # TODO: Needed for hg middleware ('lib/galaxy/webapps/tool_shed/framework/middleware/hg.py')
        tool_shed_server_wrapper = driver_util.launch_server(
            app_factory=lambda: build_shed_app(kwargs),
            webapp_factory=toolshedbuildapp.app_factory,
            galaxy_config=kwargs,
            prefix="TOOL_SHED",
            init_fast_app=init_tool_shed_fast_app,
        )
        self.server_wrappers.append(tool_shed_server_wrapper)
        tool_shed_test_host = tool_shed_server_wrapper.host
        tool_shed_test_port = tool_shed_server_wrapper.port
        log.info(f"Functional tests will be run against {tool_shed_test_host}:{tool_shed_test_port}")

        # Used by get_filename in tool shed's twilltestcase
        if "TOOL_SHED_TEST_FILE_DIR" not in os.environ:
            os.environ["TOOL_SHED_TEST_FILE_DIR"] = driver_util.TOOL_SHED_TEST_DATA

        # ---- Optionally start up a Galaxy instance ------------------------------------------------------
        if "TOOL_SHED_TEST_OMIT_GALAXY" not in os.environ:
            # Generate the shed_tool_conf.xml file.
            tool_sheds_conf_template_parser = string.Template(tool_sheds_conf_xml_template)
            tool_sheds_conf_xml = tool_sheds_conf_template_parser.safe_substitute(
                shed_url=tool_shed_test_host, shed_port=tool_shed_test_port
            )
            with open(galaxy_tool_sheds_conf_file, "w") as fh:
                fh.write(tool_sheds_conf_xml)
            # Generate the tool_sheds_conf.xml file.
            shed_tool_conf_template_parser = string.Template(shed_tool_conf_xml_template)
            shed_tool_conf_xml = shed_tool_conf_template_parser.safe_substitute(shed_tool_path=galaxy_shed_tool_path)
            with open(galaxy_shed_tool_conf_file, "w") as fh:
                fh.write(shed_tool_conf_xml)
            # Generate the migrated_tool_conf.xml file.
            migrated_tool_conf_xml = shed_tool_conf_template_parser.safe_substitute(
                shed_tool_path=galaxy_migrated_tool_path
            )
            with open(galaxy_migrated_tool_conf_file, "w") as fh:
                fh.write(migrated_tool_conf_xml)
            os.environ["GALAXY_TEST_SHED_TOOL_CONF"] = galaxy_shed_tool_conf_file
            # Generate shed_data_manager_conf.xml
            if not os.environ.get("GALAXY_SHED_DATA_MANAGER_CONF"):
                with open(galaxy_shed_data_manager_conf_file, "w") as fh:
                    fh.write(shed_data_manager_conf_xml_template)
            kwargs = dict(
                migrated_tools_config=galaxy_migrated_tool_conf_file,
                shed_data_manager_config_file=galaxy_shed_data_manager_conf_file,
                shed_tool_path=galaxy_shed_tool_path,
                tool_data_path=tool_data_path,
                tool_sheds_config_file=galaxy_tool_sheds_conf_file,
            )
            kwargs.update(
                driver_util.setup_galaxy_config(
                    galaxy_db_path,
                    use_test_file_dir=False,
                    default_install_db_merged=False,
                    default_tool_data_table_config_path=default_tool_data_table_config_path,
                    default_shed_tool_data_table_config=shed_tool_data_table_conf_file,
                    enable_tool_shed_check=True,
                    shed_tool_conf=galaxy_shed_tool_conf_file,
                    update_integrated_tool_panel=True,
                )
            )
            print("Galaxy database connection:", kwargs["database_connection"])

            # ---- Run galaxy webserver ------------------------------------------------------
            galaxy_server_wrapper = driver_util.launch_server(
                app_factory=lambda: driver_util.build_galaxy_app(kwargs),
                webapp_factory=galaxy_app_factory,
                galaxy_config=kwargs,
            )
            log.info(f"Galaxy tests will be run against {galaxy_server_wrapper.host}:{galaxy_server_wrapper.port}")
            self.server_wrappers.append(galaxy_server_wrapper)



from galaxy.tool_shed.util.container_util import (
    generate_repository_dependencies_key_for_repository,
    get_components_from_key,
    STRSEP,
)

__all__ = (
    "generate_repository_dependencies_key_for_repository",
    "get_components_from_key",
    "STRSEP",
)



import logging

log = logging.getLogger(__name__)


class Metadata:
    def __init__(self):
        self.type = None

    def get_changesets_for_setting_metadata(self, app, repository):
        repo = repository.hg_repo
        return repo.changelog

    def is_valid_for_type(self, repository, revisions_to_check=None):
        raise Exception("Unimplemented Method")


class TipOnly(Metadata):
    def __init__(self):
        self.type = None

    def get_changesets_for_setting_metadata(self, app, repository):
        repo = repository.hg_repo
        return [repo.changelog.tip()]
